# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from __future__ import annotations

import os
import subprocess
import hashlib
import uuid
from typing import Any, Final
from time import monotonic
from dataclasses import dataclass
from collections.abc import Callable

from xpra.util.child_reaper import getChildReaper
from xpra.os_util import POSIX, WIN32, gi_import
from xpra.util.io import umask_context
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool
from xpra.scripts.config import str_to_bool, parse_with_unit
from xpra.common import SizedBuffer
from xpra.net.common import PacketType, PacketElement
from xpra.util.stats import std_unit
from xpra.util.thread import start_thread
from xpra.log import Logger

GLib = gi_import("GLib")

printlog = Logger("printing")
filelog = Logger("file")

DELETE_PRINTER_FILE = envbool("XPRA_DELETE_PRINTER_FILE", True)
FILE_CHUNKS_SIZE = max(0, envint("XPRA_FILE_CHUNKS_SIZE", 65536))
MAX_CONCURRENT_FILES = max(1, envint("XPRA_MAX_CONCURRENT_FILES", 10))
PRINT_JOB_TIMEOUT = max(60, envint("XPRA_PRINT_JOB_TIMEOUT", 3600))
SEND_REQUEST_TIMEOUT = max(300, envint("XPRA_SEND_REQUEST_TIMEOUT", 3600))
CHUNK_TIMEOUT = 10 * 1000

MIMETYPE_EXTS: dict[str, str] = {
    "application/postscript": "ps",
    "application/pdf": "pdf",
    "raw": "raw",
}

DENY: Final[int] = 0
ACCEPT: Final[int] = 1  # the file / URL will be sent
OPEN: Final[int] = 2  # don't send, open on sender


def osclose(fd: int) -> None:
    try:
        os.close(fd)
    except OSError as e:
        filelog("os.close(%s)", fd, exc_info=True)
        filelog.error("Error closing file download:")
        filelog.estr(e)


def basename(filename: str) -> str:
    # we can't use os.path.basename,
    # because the remote end may have sent us a filename
    # which is using a different pathsep
    tmp = filename
    for sep in ("\\", "/", os.sep):
        i = tmp.rfind(sep) + 1
        tmp = tmp[i:]
    filename = tmp
    if WIN32:  # pragma: no cover
        # many characters aren't allowed at all on win32:
        tmp = ""
        for char in filename:
            if ord(char) < 32 or char in ("<", ">", ":", "\"", "|", "?", "*"):
                char = "_"
            tmp += char
    return tmp


def safe_open_download_file(basefilename: str, mimetype: str) -> tuple[str, int]:
    from xpra.platform.paths import get_download_dir  # pylint: disable=import-outside-toplevel
    dd = os.path.expanduser(get_download_dir())
    filename = os.path.abspath(os.path.join(dd, basename(basefilename)))
    ext = MIMETYPE_EXTS.get(mimetype)
    if ext and not filename.endswith("." + ext):
        # on some platforms (win32),
        # we want to force an extension
        # so that the file manager can display them properly when you double-click on them
        filename += "." + ext
    # make sure we use a filename that does not exist already:
    root, ext = os.path.splitext(filename)
    base = 0
    while os.path.exists(filename):
        filelog(f"cannot save file as {filename!r}: file already exists")
        base += 1
        filename = f"{root}-{base}{ext}"
    filelog("safe_open_download_file(%s, %s) will use %r", basefilename, mimetype, filename)
    flags = os.O_CREAT | os.O_RDWR | os.O_EXCL
    try:
        flags |= os.O_BINARY  # @UndefinedVariable (win32 only)
    except AttributeError:
        pass
    with umask_context(0o133):
        fd = os.open(filename, flags)
    filelog(f"using {filename=!r}, {fd=}")
    return filename, fd


@dataclass
class ReceiveChunkState:
    start: float
    fd: int
    filename: str
    mimetype: str
    printit: bool
    openit: bool
    filesize: int
    options: typedict
    digest: hashlib._Hash | None
    written: int
    cancelled: bool
    send_id: str
    timer: int
    chunk: int


@dataclass
class SendChunkState:
    start: float
    data: bytes
    chunk_size: int
    timer: int
    chunk: int


@dataclass
class SendPendingData:
    datatype: str
    url: str
    mimetype: str
    data: bytes
    filesize: int
    printit: bool
    openit: bool
    options: dict


class FileTransferAttributes:

    def __init__(self):
        self.init_attributes()

    def init_opts(self, opts, can_ask=True) -> None:
        # get the settings from a config object
        self.init_attributes(opts.file_transfer, opts.file_size_limit,
                             opts.printing, opts.open_files, opts.open_url, opts.open_command, can_ask)

    def init_attributes(self, file_transfer="no", file_size_limit="1G", printing="no",
                        open_files="no", open_url="no", open_command=None, can_ask=True) -> None:
        filelog("file transfer: init_attributes%s",
                (file_transfer, file_size_limit, printing, open_files, open_url, open_command, can_ask))

        def pask(v) -> bool:
            return v.lower() in ("ask", "auto")

        fta = pask(file_transfer)
        self.file_transfer_ask = fta and can_ask
        self.file_transfer = fta or str_to_bool(file_transfer)
        self.file_size_limit = parse_with_unit("file-size-limit", file_size_limit, "B", min_value=0) or 0
        self.file_chunks = FILE_CHUNKS_SIZE
        pa = pask(printing)
        self.printing_ask = pa and can_ask
        self.printing = pa or str_to_bool(printing)
        ofa = pask(open_files)
        self.open_files_ask = ofa and can_ask
        self.open_files = ofa or str_to_bool(open_files)
        # FIXME: command line options needed here:
        oua = pask(open_url)
        self.open_url_ask = oua and can_ask
        self.open_url = oua or str_to_bool(open_url)
        self.file_ask_timeout = SEND_REQUEST_TIMEOUT
        self.open_command = open_command
        self.files_requested: dict[str, bool] = {}
        self.files_accepted: dict[str, bool] = {}
        self.file_request_callback: dict[str, Callable[[str, int], None]] = {}
        filelog("file transfer attributes=%s", self.get_file_transfer_features())

    def get_file_transfer_features(self) -> dict[str, Any]:
        # used in hello packets,
        # duplicated with namespace (old caps to be removed in v6)
        return {
            "file-transfer": self.file_transfer,
            "file-transfer-ask": self.file_transfer_ask,
            "max-file-size": self.file_size_limit,
            "file-chunks": self.file_chunks,
            "open-files": self.open_files,
            "open-files-ask": self.open_files_ask,
            "printing": self.printing,
            "printing-ask": self.printing_ask,
            "open-url": self.open_url,
            "open-url-ask": self.open_url_ask,
            "file-ask-timeout": self.file_ask_timeout,
            # v5 onwards can use a proper namespace:
            "file": self.get_file_transfer_info(),
        }

    def get_info(self) -> dict[str, Any]:
        return self.get_file_transfer_info()

    def get_file_transfer_info(self) -> dict[str, Any]:
        # slightly different from above... for legacy reasons
        # this one is used for get_info() in a proper "file." namespace from base.py
        return {
            "enabled": self.file_transfer,
            "ask": self.file_transfer_ask,
            "size-limit": self.file_size_limit,
            "chunks": self.file_chunks,
            "open": self.open_files,
            "open-ask": self.open_files_ask,
            "open-url": self.open_url,
            "open-url-ask": self.open_url_ask,
            "printing": self.printing,
            "printing-ask": self.printing_ask,
            "ask-timeout": self.file_ask_timeout,
        }


def digest_mismatch(filename: str, digest, expected_digest) -> None:
    filelog.error(f"Error: data does not match, invalid {digest.name} file digest")
    filelog.error(f" for {filename!r}")
    filelog.error(f" received {digest.hexdigest()}")
    filelog.error(f" expected {expected_digest}")


def get_open_env() -> dict[str, str]:
    env = os.environ.copy()
    # prevent loops:
    env["XPRA_XDG_OPEN"] = "1"
    return env


class FileTransferHandler(FileTransferAttributes):
    """
        Utility class for receiving files and optionally printing them,
        used by both clients and server to share the common code and attributes
    """

    def init_attributes(self, *args) -> None:
        super().init_attributes(*args)
        self.remote_file_transfer = False
        self.remote_file_transfer_ask = False
        self.remote_printing = False
        self.remote_printing_ask = False
        self.remote_open_files = False
        self.remote_open_files_ask = False
        self.remote_open_url = False
        self.remote_open_url_ask = False
        self.remote_file_ask_timeout = SEND_REQUEST_TIMEOUT
        self.remote_file_size_limit = 0
        self.remote_file_chunks = 0
        self.pending_send_data: dict[str, SendPendingData] = {}
        self.pending_send_data_timers: dict[str, int] = {}
        self.send_chunks_in_progress: dict[str, SendChunkState] = {}
        self.receive_chunks_in_progress: dict[str, ReceiveChunkState] = {}
        self.file_descriptors: set[int] = set()

    def cleanup(self) -> None:
        for t in self.pending_send_data_timers.values():
            GLib.source_remove(t)
        self.pending_send_data_timers = {}
        for v in self.receive_chunks_in_progress.values():
            GLib.source_remove(v.timer)
        self.receive_chunks_in_progress = {}
        for x in tuple(self.file_descriptors):
            try:
                os.close(x)
            except OSError as e:
                filelog(f"os.close({x}) {e}")
        self.file_descriptors = set()
        self.init_attributes()

    def parse_file_transfer_caps(self, c) -> None:
        fc = typedict(c.dictget("file") or {})
        filelog("parse_file_transfer_caps: %s", fc)
        # v5 with "file" namespace:
        self.remote_file_transfer = fc.boolget("enabled")
        self.remote_file_transfer_ask = fc.boolget("ask")
        self.remote_printing = fc.boolget("printing")
        self.remote_printing_ask = fc.boolget("printing-ask")
        self.remote_open_files = fc.boolget("open")
        self.remote_open_files_ask = fc.boolget("open-ask")
        self.remote_open_url = fc.boolget("open-url")
        self.remote_open_url_ask = fc.boolget("open-url-ask")
        self.remote_file_ask_timeout = fc.intget("ask-timeout")
        self.remote_file_size_limit = fc.intget("max-file-size") or fc.intget("size-limit")
        self.remote_file_chunks = max(0, fc.intget("chunks"))
        self.dump_remote_caps()

    def dump_remote_caps(self) -> None:
        filelog("file transfer remote caps:")
        filelog(" file-transfer=%-5s   (ask=%s)", self.remote_file_transfer, self.remote_file_transfer_ask)
        filelog(" printing=%-5s        (ask=%s)", self.remote_printing, self.remote_printing_ask)
        filelog(" open-files=%-5s      (ask=%s)", self.remote_open_files, self.remote_open_files_ask)
        filelog(" open-url=%-5s        (ask=%s)", self.remote_open_url, self.remote_open_url_ask)
        filelog(" file-size-limit=%s", self.remote_file_size_limit)

    def get_info(self) -> dict[str, Any]:
        info = super().get_info()
        info["remote"] = {
            "file-transfer": self.remote_file_transfer,
            "file-transfer-ask": self.remote_file_transfer_ask,
            "file-size-limit": self.remote_file_size_limit,
            "file-chunks": self.remote_file_chunks,
            "open-files": self.remote_open_files,
            "open-files-ask": self.remote_open_files_ask,
            "open-url": self.remote_open_url,
            "open-url-ask": self.remote_open_url_ask,
            "printing": self.remote_printing,
            "printing-ask": self.remote_printing_ask,
            "file-ask-timeout": self.remote_file_ask_timeout,
        }
        return info

    def _check_chunk_receiving(self, chunk_id: str, chunk_no: int) -> None:
        chunk_state = self.receive_chunks_in_progress.get(chunk_id)
        filelog("_check_chunk_receiving(%s, %s) chunk_state=%s", chunk_id, chunk_no, chunk_state)
        if not chunk_state:
            # transfer not found
            return
        if chunk_state.cancelled:
            # transfer has been cancelled
            return
        chunk_state.timer = 0  # this timer has been used
        if chunk_state.chunk == chunk_no:
            filelog.error(f"Error: chunked file transfer f{chunk_id} timed out")
            self.receive_chunks_in_progress.pop(chunk_id, None)

    def cancel_download(self, send_id: str, message="Cancelled") -> None:
        filelog("cancel_download(%s, %s)", send_id, message)
        for chunk_id, chunk_state in dict(self.receive_chunks_in_progress).items():
            if chunk_state.send_id == send_id:
                self.cancel_file(chunk_id, message)
                return
        filelog.error(f"Error: cannot cancel download {send_id!r}, entry not found!")

    def cancel_file(self, chunk_id: str, message: str, chunk: int = 0) -> None:
        filelog("cancel_file%s", (chunk_id, message, chunk))
        chunk_state = self.receive_chunks_in_progress.get(chunk_id)
        if chunk_state:
            # mark it as cancelled:
            chunk_state.cancelled = True
            timer = chunk_state.timer
            if timer:
                chunk_state.timer = 0
                GLib.source_remove(timer)
            osclose(chunk_state.fd)

            # remove this transfer after a little while,
            # so in-flight packets won't cause errors

            def clean_receive_state() -> bool:
                self.receive_chunks_in_progress.pop(chunk_id, None)
                return False

            GLib.timeout_add(20000, clean_receive_state)
            filename = chunk_state.filename
            try:
                os.unlink(filename)
            except OSError as e:
                filelog(f"os.unlink({filename})", exc_info=True)
                filelog.error("Error: failed to delete temporary download file")
                filelog.error(f" {filename!r} : {e}")
        self.send("ack-file-chunk", chunk_id, False, message, chunk)

    def _process_send_file_chunk(self, packet: PacketType) -> None:
        chunk_id = str(packet[1])
        chunk = int(packet[2])
        file_data: SizedBuffer = packet[3]
        has_more = bool(packet[4])
        # if len(file_data)<1024:
        #    from xpra.util.str_fn import hexstr
        #    filelog.warn("file_data=%s", hexstr(file_data))
        # filelog(f"file_data={len(file_data)} {type(file_data)}")
        filelog(f"file_data={len(file_data)} {type(file_data)}")
        filelog("_process_send_file_chunk%s", (chunk_id, chunk, f"{len(file_data)} bytes", has_more))
        chunk_state = self.receive_chunks_in_progress.get(chunk_id)

        def progress(position: int, error="") -> None:
            s_elapsed = monotonic() - chunk_state.start
            self.transfer_progress_update(False, chunk_state.send_id, s_elapsed, position, chunk_state.filesize, error)

        def error(message: str) -> None:
            filelog.error(f"Error: {message}")
            self.cancel_file(chunk_id, message, chunk)
            if chunk_state:
                if chunk_state.fd:
                    osclose(chunk_state.fd)
                filename = chunk_state.filename
                if filename and os.path.exists(filename):
                    try:
                        os.unlink(filename)
                    except OSError:
                        filelog.error(f"Error: failed to delete uploaded file {filename!r}")
            progress(-1, message)

        if not chunk_state:
            error(f"file transfer id {chunk_id!r} not found")
            return

        if chunk_state.cancelled:
            filelog("got chunk for a cancelled file transfer, ignoring it")
            return

        fd = chunk_state.fd
        if chunk_state.chunk + 1 != chunk:
            error("chunk number mismatch")
            return

        # update chunk number:
        chunk_state.chunk = chunk
        try:
            os.write(fd, file_data)
            if chunk_state.digest:
                chunk_state.digest.update(file_data)
            chunk_state.written += len(file_data)
        except OSError as e:
            error(f"cannot write file chunk: {e}")
            return

        if chunk_state.written > chunk_state.filesize:
            error("more data received than specified in the file size!")
            return

        self.send("ack-file-chunk", chunk_id, True, "", chunk)
        if chunk_state.cancelled:
            # check again if the transfer has been cancelled
            filelog("got chunk for a cancelled file transfer, ignoring it")
            return

        if has_more:
            progress(chunk_state.written)
            if chunk_state.timer:
                GLib.source_remove(chunk_state.timer)
            # remote end will send more after receiving the ack
            chunk_state.timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
            pct = round(100 * chunk_state.written / chunk_state.filesize)
            filelog(f"waiting for the next chunk, got {chunk_state.written:8} of {chunk_state.filesize:8}: {pct:3}%")
            return

        # we have received all the packets
        self.receive_chunks_in_progress.pop(chunk_id, None)
        osclose(fd)
        filename = chunk_state.filename
        options = chunk_state.options
        filelog(f"file {filename!r} complete")
        if chunk_state.digest:
            expected_digest = options.strget(chunk_state.digest.name)  # ie: "sha256"
            if expected_digest and chunk_state.digest.hexdigest() != expected_digest:
                digest_mismatch(filename, chunk_state.digest, expected_digest)
                error("checksum mismatch")
                return
            filelog("%s digest matches: %s", chunk_state.digest.name, expected_digest)

        # check file size and digest then process it:
        if chunk_state.written != chunk_state.filesize:
            error(f"expected a file of {chunk_state.filesize} bytes, got {chunk_state.written}")
            return

        progress(chunk_state.written)
        elapsed = monotonic() - chunk_state.start
        filelog("%i bytes received in %i chunks, took %ims", chunk_state.filesize, chunk, elapsed * 1000)
        self.process_downloaded_file(filename, chunk_state.mimetype,
                                     chunk_state.printit, chunk_state.openit, chunk_state.filesize, options)

    def accept_data(self, send_id: str, dtype, basefilename: str,
                    printit: bool, openit: bool) -> tuple[bool, bool, bool]:
        # subclasses should check the flags,
        # and if ask is True, verify they have accepted this specific send_id
        filelog("accept_data%s", (send_id, dtype, basefilename, printit, openit))
        filelog("accept_data: printing=%s, printing-ask=%s",
                self.printing, self.printing_ask)
        filelog("accept_data: file-transfer=%s, file-transfer-ask=%s",
                self.file_transfer, self.file_transfer_ask)
        filelog("accept_data: open-files=%s, open-files-ask=%s",
                self.open_files, self.open_files_ask)
        req = self.files_accepted.pop(send_id, None)
        filelog("accept_data: files_accepted[%s]=%s", send_id, req)
        if req is not None:
            return True, False, req
        if printit:
            if not self.printing or self.printing_ask:
                printit = False
        elif not self.file_transfer or self.file_transfer_ask:
            return False, False, False
        if openit and (not self.open_files or self.open_files_ask):
            # we can't ask in this implementation,
            # so deny the request to open it:
            openit = False
        return True, printit, openit

    def _process_send_file(self, packet: PacketType) -> None:
        # the remote end is sending us a file
        start = monotonic()
        basefilename = str(packet[1])
        mimetype = str(packet[2])
        printit = bool(packet[3])
        openit = bool(packet[4])
        filesize = int(packet[5])
        file_data: SizedBuffer = packet[6]
        options: typedict = typedict(packet[7])
        send_id = ""
        if len(packet) >= 9:
            send_id = str(packet[8])
        chunk_id = options.strget("file-chunk-id")

        def cancel(message: str) -> None:
            log.error(f"Error: {message}")
            if chunk_id:
                self.cancel_file(chunk_id, message, chunk)

        if filesize <= 0:
            cancel(f"invalid file size {filesize} for {basefilename!r}")
            return
        if filesize > self.file_size_limit:
            fstr = std_unit(filesize)
            limstr = std_unit(self.file_size_limit)
            cancel(f"file {basefilename!r} is too large: {fstr}B, the file size limit is {limstr}B")
            return

        args = (send_id, "file", basefilename, printit, openit)
        acceptit, printit, openit = self.accept_data(*args)
        filelog("%s%s=%s", self.accept_data, args, (acceptit, printit, openit))
        if not acceptit:
            ftype = "printing" if printit else "transfer"
            cancel(f"{ftype} rejected for file {basefilename!r}")
            return

        # accept_data() method can override the flags:
        if printit:
            log = printlog
            assert self.printing
        else:
            log = filelog
            assert self.file_transfer
        log("receiving file: %s",
            (basefilename, mimetype, printit, openit, filesize, f"{len(file_data)} bytes", options))

        try:
            filename, fd = safe_open_download_file(basefilename, mimetype)
        except OSError as e:
            log("cannot save file %s / %s", basefilename, mimetype, exc_info=True)
            cancel(f"failed to save downloaded file: {e}")
            return
        self.file_descriptors.add(fd)

        # from now on, call error() so we also close / cleanup the temporary file:
        def error(message: str) -> None:
            cancel(message)
            if fd:
                try:
                    os.close(fd)
                except OSError:
                    log.error(f"Error: failed to close {fd=} used for file transfer")
                else:
                    try:
                        self.file_descriptors.remove(fd)
                    except KeyError:
                        pass

        digest: hashlib._Hash | None = None
        for hash_fn in ("sha512", "sha384", "sha256", "sha224", "sha1"):
            if options.strget(hash_fn):
                digest = getattr(hashlib, hash_fn)()
                break
        if chunk_id:
            chunk = 0
            nfiles = len(self.receive_chunks_in_progress)
            if nfiles >= MAX_CONCURRENT_FILES:
                error(f"too many file transfers in progress: {nfiles}")
                return
            timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_receiving, chunk_id, chunk)
            self.receive_chunks_in_progress[chunk_id] = ReceiveChunkState(monotonic(),
                                                                          fd, filename, mimetype,
                                                                          printit, openit, filesize,
                                                                          options, digest, 0, False, send_id,
                                                                          timer, chunk)
            self.send("ack-file-chunk", chunk_id, True, b"", chunk)
            return
        # not chunked, full file:
        if not file_data:
            error("no file data")
            return
        if len(file_data) != filesize:
            error(f"file {basefilename!r} size: received {len(file_data)} bytes, expected {filesize} bytes")
            return
        # check digest if present:
        if digest:
            digest.update(file_data)
            expected_digest = options.strget(digest.name)  # ie: "sha256"
            if expected_digest and digest.hexdigest() != expected_digest:
                digest_mismatch(basefilename, digest, expected_digest)
                error("checksum mismatch")
                return
            log("%s digest matches: %s", digest.name, expected_digest)
        try:
            os.write(fd, file_data)
        except OSError as e:
            error(f"failed to write file data: {e}")
            return
        finally:
            osclose(fd)
        self.transfer_progress_update(False, send_id, monotonic() - start, filesize, filesize, None)
        self.process_downloaded_file(filename, mimetype, printit, openit, filesize, options)

    def process_downloaded_file(self, filename: str, mimetype: str,
                                printit: bool, openit: bool, filesize: int, options) -> None:
        sizestr = std_unit(filesize)
        ftype = mimetype or "temporary"
        extra = " for printing" if printit else ""
        filelog.info(f"downloaded {sizestr} bytes to {ftype} file{extra}:")
        filelog.info(f" {filename!r}")
        # some file requests may have a custom callback
        # (ie: bug report tool will just include the file)
        rf = options.tupleget("request-file")
        if rf and len(rf) >= 2:
            argf = rf[0]
            cb = self.file_request_callback.pop(argf, None)
            if cb:
                cb(filename, filesize)
                return
        if printit or openit:
            t = start_thread(self.do_process_downloaded_file, "process-download", daemon=False,
                             args=(filename, mimetype, printit, openit, filesize, options))
            filelog("started process-download thread: %s", t)

    def do_process_downloaded_file(self, filename: str, mimetype: str, printit: bool, openit: bool, filesize: int,
                                   options: typedict) -> None:
        filelog("do_process_downloaded_file%s", (filename, mimetype, printit, openit, filesize, options))
        if printit:
            self._print_file(filename, mimetype, options)
            return
        if openit:
            if not self.open_files:
                filelog.warn("Warning: opening files automatically is disabled,")
                filelog.warn(" ignoring uploaded file:")
                filelog.warn(" '%s'", filename)
                return
            self._open_file(filename)

    def _print_file(self, filename: str, mimetype: str, options: typedict) -> None:
        printlog("print_file%s", (filename, mimetype, options))
        printer = options.strget("printer")
        title = options.strget("title")
        copies = options.intget("copies", 1)
        if title:
            printlog.info(" sending '%s' to printer '%s'", title, printer)
        else:
            printlog.info(" sending to printer '%s'", printer)
        from xpra.platform.printing import print_files, printing_finished, get_printers
        printers = get_printers()

        def delfile() -> None:
            if DELETE_PRINTER_FILE:
                try:
                    os.unlink(filename)
                except OSError:
                    printlog("failed to delete print job file '%s'", filename)

        if not printer:
            printlog.error("Error: the printer name is missing")
            printlog.error(" printers available: %s", csv(printers.keys()) or "none")
            delfile()
            return
        if printer not in printers:
            printlog.error(f"Error: printer {printer!r} does not exist!")
            printlog.error(" printers available: %s", csv(printers.keys()) or "none")
            delfile()
            return
        try:
            job_options = options.dictget("options", {})
            job_options["copies"] = copies
            job = print_files(printer, [filename], title, job_options)
        except Exception as e:
            printlog("print_files%s", (printer, [filename], title, options), exc_info=True)
            printlog.error("Error: cannot print file '%s'", os.path.basename(filename))
            printlog.estr(e)
            delfile()
            return
        printlog(f"printing {filename!r}, job={job}")
        if job <= 0:
            printlog(f"printing failed and returned {job}")
            delfile()
            return
        start = monotonic()

        def check_printing_finished() -> bool:
            done = printing_finished(job)
            printlog("printing_finished(%s)=%s", job, done)
            if done:
                delfile()
                return False
            if monotonic() - start >= PRINT_JOB_TIMEOUT:
                printlog.warn("Warning: print job %s timed out", job)
                delfile()
                return False
            return True  # try again..

        if check_printing_finished():
            # check every 10 seconds:
            GLib.timeout_add(10000, check_printing_finished)

    def _open_file(self, url: str) -> None:
        filelog("_open_file(%s)", url)
        self.exec_open_command(url)

    def _open_url(self, url: str) -> None:
        filelog("_open_url(%s)", url)
        if POSIX:
            # we can't use webbrowser,
            # because this will use "xdg-open" from the $PATH
            # which may point back to us!
            self.exec_open_command(url)
        else:
            import webbrowser  # pylint: disable=import-outside-toplevel
            webbrowser.open_new_tab(url)

    def exec_open_command(self, url: str) -> None:
        filelog("exec_open_command(%s)", url)
        try:
            import shlex  # pylint: disable=import-outside-toplevel
            command = shlex.split(self.open_command) + [url]
        except ImportError as e:
            filelog("exec_open_command(%s) no shlex: %s", url, e)
            command = self.open_command.split(" ")
        filelog("exec_open_command(%s) command=%s", url, command)
        try:
            # pylint: disable=consider-using-with
            proc = subprocess.Popen(command, env=get_open_env(), shell=WIN32)
        except Exception as e:
            filelog("exec_open_command(%s)", url, exc_info=True)
            filelog.error("Error: cannot open '%s': %s", url, e)
            return
        filelog("exec_open_command(%s) Popen(%s)=%s", url, command, proc)

        def open_done(*_args) -> None:
            returncode = proc.poll()
            filelog("open_file: command %s has ended, returncode=%s", command, returncode)
            if returncode != 0:
                filelog.warn("Warning: failed to open the downloaded content")
                filelog.warn(" '%s' returned %s", " ".join(command), returncode)

        cr = getChildReaper()
        cr.add_process(proc, f"Open file {url}", command, True, True, open_done)

    def file_size_warning(self, action: str, location: str, basefilename: str, filesize: int, limit: int) -> None:
        filelog.warn("Warning: cannot %s the file '%s'", action, basefilename)
        filelog.warn(" this file is too large: %sB", std_unit(filesize))
        filelog.warn(" the %s file size limit is %sB", location, std_unit(limit))

    def check_file_size(self, action: str, filename: str, filesize: int) -> bool:
        basefilename = os.path.basename(filename)
        if filesize > self.file_size_limit:
            self.file_size_warning(action, "local", basefilename, filesize, self.file_size_limit)
            return False
        if filesize > self.remote_file_size_limit:
            self.file_size_warning(action, "remote", basefilename, filesize, self.remote_file_size_limit)
            return False
        return True

    def send_request_file(self, filename: str, openit: bool = True):
        self.send("request-file", filename, openit)
        self.files_requested[filename] = openit

    def _process_open_url(self, packet: PacketType) -> None:
        send_id = str(packet[2])
        url = str(packet[1])
        if not self.open_url:
            filelog.warn("Warning: received a request to open URL '%s'", url)
            filelog.warn(" but opening of URLs is disabled")
            return
        if self.open_url_ask:
            acceptit, _printit, _openit = self.accept_data(send_id, "url", url, False, True)
            if not acceptit:
                filelog("url '%s' not accepted", url)
                return
        self._open_url(url)

    def send_open_url(self, url: str) -> bool:
        if not self.remote_open_url:
            filelog.warn("Warning: remote end does not accept URLs")
            return False
        if self.remote_open_url_ask:
            # ask the client if it is OK to send
            return bool(self.send_data_request("open", "url", url))
        self.do_send_open_url(url)
        return True

    def do_send_open_url(self, url: str, send_id: str = "") -> None:
        self.send("open-url", url, send_id)

    def send_file(self, filename: str, mimetype: str, data: SizedBuffer, filesize=0,
                  printit=False, openit=False, options=None) -> bool:
        if printit:
            if not self.printing:
                printlog.warn("Warning: printing is not enabled for %s", self)
                return False
            if not self.remote_printing:
                printlog.warn("Warning: remote end does not support printing")
                return False
            logger = printlog
            ask = self.remote_printing_ask
            action = "print"
        else:
            if not self.file_transfer:
                filelog.warn("Warning: file transfers are not enabled for %s", self)
                return False
            if not self.remote_file_transfer:
                filelog.warn("Warning: remote end does not support file transfers")
                return False
            logger = filelog
            ask = self.remote_file_transfer_ask
            action = "upload"
            if openit:
                if not self.remote_open_files:
                    logger.info("opening the file after transfer is disabled on the remote end")
                    logger.info(" sending only, the file will need to be opened manually")
                    openit = False
                    action = "upload"
                else:
                    ask |= self.remote_open_files_ask
                    action = "open"
        assert len(data) >= filesize, "data is smaller then the given file size!"
        data = data[:filesize]  # gio may null terminate it
        logger("send_file%s action=%s, ask=%s",
               (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options), action, ask)
        self.dump_remote_caps()
        if not self.check_file_size(action, filename, filesize):
            return False
        if ask:
            return bool(self.send_data_request(action, "file", filename, mimetype, data, filesize,
                                               printit, openit, options))
        send_id = uuid.uuid4().hex
        self.do_send_file(filename, mimetype, data, filesize, printit, openit, options, send_id)
        return True

    def send_data_request(self, action: str, dtype: str, url: str, mimetype="", data=b"", filesize=0,
                          printit=False, openit=True, options=None) -> str:
        send_id = uuid.uuid4().hex
        if len(self.pending_send_data) >= MAX_CONCURRENT_FILES:
            filelog.warn("Warning: %s dropped", action)
            filelog.warn(" %i transfers already waiting for a response", len(self.pending_send_data))
            return ""
        spd = SendPendingData(dtype, url, mimetype, data, filesize, printit, openit, options or {})
        self.pending_send_data[send_id] = spd
        delay = self.remote_file_ask_timeout * 1000
        self.pending_send_data_timers[send_id] = GLib.timeout_add(delay, self.send_data_ask_timeout, send_id)
        filelog(f"sending data request for {dtype} {url!r} with send-id {send_id!r}")
        self.send("send-data-request", dtype, send_id, url, mimetype, filesize, printit, openit, options or {})
        return send_id

    def _process_send_data_request(self, packet: PacketType) -> None:
        dtype, send_id, url, _, filesize, printit, openit = packet[1:8]
        options = {}
        if len(packet) >= 9:
            options = packet[8]
        # filenames and url are always sent encoded as utf8:
        self.do_process_send_data_request(dtype, send_id, url, _, filesize, printit, openit, typedict(options))

    def do_process_send_data_request(self, dtype: str, send_id: str, url: str, _, filesize: int,
                                     printit: bool, openit: bool, options: typedict) -> None:
        filelog(f"do_process_send_data_request: {send_id=}, {url=}, {printit=}, {openit=}, {options=}")

        def cb_answer(accept: bool) -> None:
            filelog("accept%s=%s", (url, printit, openit), accept)
            self.send("send-data-response", send_id, accept)

        # could be a request we made:
        # (in which case we can just accept it without prompt)
        rf = options.tupleget("request-file")
        if rf and len(rf) >= 2:
            argf, openit = rf[:2]
            openit = self.files_requested.pop(argf, None)
            if openit is not None:
                self.files_accepted[send_id] = openit
                cb_answer(True)
                return
        if dtype == "file":
            if not self.file_transfer:
                cb_answer(False)
                return
            url = os.path.basename(url)
            if printit:
                ask = self.printing_ask
            elif openit:
                ask = self.file_transfer_ask or self.open_files_ask
            else:
                ask = self.file_transfer_ask
        elif dtype == "url":
            if not self.open_url:
                cb_answer(False)
                return
            ask = self.open_url_ask
        else:
            filelog.warn("Warning: unknown data request type '%s'", dtype)
            cb_answer(False)
            return
        if not ask:
            filelog.warn("Warning: received a send-data request for a %s,", dtype)
            filelog.warn(" but authorization is not required by the client")
            # fail it because if we responded with True,
            # it would fail later when we don't find this send_id in our accepted list
            cb_answer(False)
            return
        self.ask_data_request(cb_answer, send_id, dtype, url, filesize, printit, openit)

    def ask_data_request(self, cb_answer: Callable[[bool], None],
                         send_id: str, dtype: str, url: str, filesize: int,
                         printit: bool, openit: bool) -> None:
        # subclasses may prompt the user here instead
        filelog("ask_data_request%s", (send_id, dtype, url, filesize, printit, openit))
        v = self.accept_data(send_id, dtype, url, printit, openit)
        cb_answer(v[0])

    def _process_send_data_response(self, packet: PacketType) -> None:
        send_id = str(packet[1])
        accept = int(packet[2])
        filelog("process send-data-response: send_id=%s, accept=%s", send_id, accept)
        timer = self.pending_send_data_timers.pop(send_id, None)
        if timer:
            GLib.source_remove(timer)
        try:
            spd: SendPendingData = self.pending_send_data[send_id]
        except KeyError:
            filelog.warn(f"Warning: cannot find send-file entry for {send_id!r}")
            return
        if accept == DENY:
            filelog.info("the request to send %s '%s' has been denied", spd.datatype, spd.url)
            return
        if accept not in (ACCEPT, OPEN):
            raise ValueError(f"unknown value for send-data response: {accept!r}")
        if spd.datatype == "file":
            if accept == ACCEPT:
                self.do_send_file(spd.url, spd.mimetype, spd.data, spd.filesize,
                                  spd.printit, spd.openit, spd.options, send_id)
            else:
                assert spd.openit and accept == OPEN
                # try to open at this end:
                self._open_file(spd.url)
        elif spd.datatype == "url":
            if accept == ACCEPT:
                self.do_send_open_url(spd.url, send_id)
            else:
                assert accept == OPEN
                # open it at this end:
                self._open_url(spd.url)
        else:
            filelog.error("Error: unknown datatype '%s'", spd.datatype)

    def send_data_ask_timeout(self, send_id: str) -> bool:
        self.pending_send_data_timers.pop(send_id, None)
        spd = self.pending_send_data.pop(send_id, None)
        if not spd:
            filelog.warn(f"Warning: send timeout, id {send_id!r} not found!")
            return False
        filelog.warn("Warning: failed to %s file '%s',", ["send", "print"][spd.printit], spd.url)
        filelog.warn(" the send approval request timed out")
        return False

    def do_send_file(self, filename: str, mimetype: str, data, filesize: int = 0,
                     printit: bool = False, openit: bool = False, options=None, send_id: str = "") -> bool:
        if printit:
            action = "print"
            logger = printlog
        else:
            action = "upload"
            logger = filelog
        logger("do_send_file%s", (filename, mimetype, type(data), f"{filesize} bytes", printit, openit, options))
        if not self.check_file_size(action, filename, filesize):
            return False
        h = hashlib.sha256()
        h.update(data)
        absfile = os.path.abspath(filename)
        filelog(f"sha256 digest({absfile})=%s", h.hexdigest())
        options = options or {}
        options["sha256"] = h.hexdigest()
        chunk_size = min(self.file_chunks, self.remote_file_chunks)
        if 0 < chunk_size < filesize:
            in_progress = len(self.send_chunks_in_progress)
            if in_progress >= MAX_CONCURRENT_FILES:
                raise RuntimeError(f"too many file transfers in progress: {in_progress}")
            # chunking is supported and the file is big enough
            chunk_id = uuid.uuid4().hex
            options["file-chunk-id"] = chunk_id
            # timer to check that the other end is requesting more chunks:
            chunk_no = 0
            timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk_no)
            self.send_chunks_in_progress[chunk_id] = SendChunkState(monotonic(), data, chunk_size, timer, chunk_no)
            cdata = b""
            filelog("using chunks, sending initial file-chunk-id=%s, for chunk size=%s",
                    chunk_id, chunk_size)
        else:
            # send everything now:
            cdata = self.compressed_wrapper("file-data", data)
            assert len(cdata) <= filesize  # compressed wrapper ensures this is true
            filelog("sending full file: %i bytes (chunk size=%i)", filesize, chunk_size)
        basefilename = os.path.basename(filename)
        self.send("send-file", basefilename, mimetype, printit, openit, filesize, cdata, options, send_id)
        return True

    def _check_chunk_sending(self, chunk_id: str, chunk_no: int) -> None:
        chunk_state = self.send_chunks_in_progress.get(chunk_id)
        filelog("_check_chunk_sending(%s, %s) chunk_state found: %s", chunk_id, chunk_no, bool(chunk_state))
        if not chunk_state:
            # transfer already removed
            return
        chunk_state.timer = 0  # timer has fired
        if chunk_state.chunk == chunk_no:
            filelog.error(f"Error: chunked file transfer {chunk_id} timed out")
            filelog.error(f" on chunk {chunk_no}")
            self.cancel_sending(chunk_id)

    def cancel_sending(self, chunk_id: str) -> None:
        chunk_state = self.send_chunks_in_progress.pop(chunk_id, None)
        filelog("cancel_sending(%s) chunk state found: %s", chunk_id, bool(chunk_state))
        if not chunk_state:
            return
        timer = chunk_state.timer
        if timer:
            chunk_state.timer = 0
            GLib.source_remove(timer)

    def _process_ack_file_chunk(self, packet: PacketType) -> None:
        # the other end received our send-file or send-file-chunk,
        # send some more file data
        filelog("ack-file-chunk: %s", packet[1:])
        chunk_id = str(packet[1])
        state = bool(packet[2])
        error_message = str(packet[3])
        chunk = int(packet[4])
        if not state:
            filelog.info("the remote end is cancelling the file transfer:")
            filelog.info(" %s", error_message)
            self.cancel_sending(chunk_id)
            return
        chunk_state = self.send_chunks_in_progress.get(chunk_id)
        if not chunk_state:
            filelog.error(f"Error: cannot find the file transfer id {chunk_id!r}")
            return
        if chunk_state.chunk != chunk:
            filelog.error("Error: chunk number mismatch (%i vs %i)", chunk_state.chunk, chunk)
            self.cancel_sending(chunk_id)
            return
        chunk_size = chunk_state.chunk_size
        if not chunk_state.data:
            # all sent!
            elapsed = monotonic() - chunk_state.start
            if elapsed > 0:
                bs = chunk * chunk_size / elapsed
            else:
                bs = 9999 * 1000 * 1000
            filelog("%i chunks of %i bytes sent in %ims (%sB/s)",
                    chunk, chunk_size, elapsed * 1000, std_unit(bs))
            self.cancel_sending(chunk_id)
            return
        assert chunk_size > 0
        # carve out another chunk:
        cdata = self.compressed_wrapper("file-data", chunk_state.data[:chunk_size])
        chunk_state.data = chunk_state.data[chunk_size:]
        chunk += 1
        chunk_state.chunk = chunk
        if chunk_state.timer:
            GLib.source_remove(chunk_state.timer)
        chunk_state.timer = GLib.timeout_add(CHUNK_TIMEOUT, self._check_chunk_sending, chunk_id, chunk)
        self.send("send-file-chunk", chunk_id, chunk, cdata, bool(chunk_state.data))

    def send(self, packet_type: str, *parts: PacketElement) -> None:
        raise NotImplementedError()

    def compressed_wrapper(self, datatype, data, level=5):
        raise NotImplementedError()

    def transfer_progress_update(self, send=True, transfer_id: str = "", elapsed=0.0, position=0, total=0,
                                 error=None) -> None:
        # this method is overridden in the gtk client:
        filelog("transfer_progress_update%s", (send, transfer_id, elapsed, position, total, error))



# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# authentication from a file containing just the password

from xpra.net.digest import verify_digest
from xpra.auth.file_auth_base import FileAuthenticatorBase, log
from xpra.util.objects import typedict
from xpra.util.str_fn import obsc


class Authenticator(FileAuthenticatorBase):

    def authenticate_hmac(self, caps: typedict) -> bool:
        challenge_response = caps.bytesget("challenge_response")
        client_salt = caps.strget("challenge_client_salt")
        log(f"file_auth.authenticate_hmac challenge-response={challenge_response!r}, client-salt={client_salt!r}")
        if not self.salt:
            log.error("Error: illegal challenge response received - salt cleared or unset")
            return False
        salt = self.get_response_salt(client_salt)
        password = self.get_password()
        log("authenticate_hmac(..) get_password()=" + obsc(password))
        if not password:
            log.warn("Warning: authentication failed")
            log.warn(f" no password for {self.username!r} in {self.password_filename!r}")
            return False
        verified = verify_digest(self.digest, password, salt, challenge_response)
        log(f"authenticate_hmac(..) {verified=}")
        if not verified:
            log.warn(f"Warning: {self.digest!r} challenge for {self.username!r} does not match")
        return verified

    def parse_filedata(self, data: str) -> str:
        return data

    def get_password(self) -> str:
        password = super().get_password()
        if not password:
            return password
        if password.find("\n") >= 0 or password.find("\r") >= 0:
            log.warn("Warning: newline found in password data")
            log.warn(" this is usually a mistake")
        return password

    def __repr__(self):
        return "password file"



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
from importlib import import_module


def main(argv) -> int:
    # pylint: disable=import-outside-toplevel
    from xpra.util.str_fn import print_nested_dict
    from xpra.util.env import numpy_import_context
    from xpra.platform import program_context
    with program_context("CUDA-Info", "CUDA Info"):
        from xpra.log import Logger, enable_color, consume_verbose_argv
        enable_color()
        consume_verbose_argv(argv, "cuda")
        log = Logger("cuda")
        with numpy_import_context("CUDA: info"):
            for component in (
                "xpra.codecs",
                "xpra.codecs.nvidia",
                "xpra.codecs.nvidia.cuda",
                "pycuda",
                "pycuda.driver",
                "xpra.codecs.nvidia.cuda.context",
            ):
                try:
                    module = import_module(component)
                except ImportError as e:
                    log.error(f"Error: the `{component}` component failed to load")
                    log.estr(e)
                    if component == "pycuda.driver":
                        log.error(" this usually happens if the CUDA library is not installed or not found")
                    return 1
                else:
                    log.debug(f"loaded {component}={module}")
        from xpra.codecs.nvidia.cuda import context
        pycuda_info = context.get_pycuda_info()
        log.info("pycuda_info")
        print_nested_dict(pycuda_info, print_fn=log.info)
        log.info("cuda_info")
        print_nested_dict(context.get_cuda_info(), print_fn=log.info)
        log.info("preferences:")
        print_nested_dict(context.get_prefs(), print_fn=log.info)
        log.info("device automatically selected:")
        log.info(" %s", context.device_info(context.select_device()[1]))
    return 0


if __name__ == "__main__":
    v = main(sys.argv)
    sys.exit(v)



# This file is part of Xpra.
# Copyright (C) 2019 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.platform import platform_import

from xpra.server.proxy.server import ProxyServer

assert ProxyServer

platform_import(globals(), "proxy_server", False, "ProxyServer")



# This file is part of Xpra.
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

# According to ICCCM 2.8/4.3, a window manager for screen N is a client which
# acquires the selection WM_S<N>.  If another client already has this
# selection, we can either abort or steal it.  Once we have it, if someone
# else steals it, then we should exit.

import sys
from typing import Final
from struct import unpack, calcsize

from xpra.gtk.error import xsync
from xpra.gtk.gobject import no_arg_signal, one_arg_signal
from xpra.x11.bindings.window import constants, X11WindowBindings
from xpra.x11.gtk.bindings import add_event_receiver, remove_event_receiver, get_xatom, get_pywindow
from xpra.x11.gtk.prop import prop_set
from xpra.exit_codes import ExitCode
from xpra.util.env import envint
from xpra.os_util import gi_import
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")
GObject = gi_import("GObject")

log = Logger("x11", "util")

SELECTION_EXIT_TIMEOUT = envint("XPRA_SELECTION_EXIT_TIMEOUT", 20)

StructureNotifyMask: Final[int] = constants["StructureNotifyMask"]
XNone: Final[int] = constants["XNone"]


class AlreadyOwned(Exception):
    pass


class ManagerSelection(GObject.GObject):
    __gsignals__ = {
        "selection-lost": no_arg_signal,
        "x11-destroy-event": one_arg_signal,
    }

    def __str__(self):  # pylint: disable=arguments-differ
        return "ManagerSelection(%s)" % self.atom

    def __init__(self, selection: str):
        super().__init__()
        self.atom = selection
        atom = Gdk.Atom.intern(selection, False)
        self.clipboard = Gtk.Clipboard.get(atom)
        self.xid: int = 0
        self.exit_timer: int = 0

    def _owner(self) -> int:
        return X11WindowBindings().XGetSelectionOwner(self.atom)

    def owned(self) -> bool:
        """Returns True if someone owns the given selection."""
        return self._owner() != XNone

    # If the selection is already owned, then raise AlreadyOwned rather
    # than stealing it.
    IF_UNOWNED = "if_unowned"
    # If the selection is already owned, then steal it, and then block until
    # the previous owner has signaled that they are done cleaning up.
    FORCE = "force"
    # If the selection is already owned, then steal it and return immediately.
    # Created for the use of tests.
    FORCE_AND_RETURN = "force_and_return"

    def acquire(self, when) -> None:
        old_owner = self._owner()
        if when is self.IF_UNOWNED and old_owner != XNone:
            raise AlreadyOwned

        # we can only set strings with GTK3,
        # we should try to be compliant with ICCCM version 2.0 (see section 4.3)
        # and use this format instead:
        # outdata.set("INTEGER", 32, pack("@ii", 2, 0))
        thestring = "VERSION"
        self.clipboard.set_text(thestring, len(thestring))

        # Having acquired the selection, we have to announce our existence
        # (ICCCM 2.8, still).  The details here probably don't matter too
        # much; I've never heard of an app that cares about these messages,
        # and metacity actually gets the format wrong in several ways (no
        # MANAGER or owner_window atoms).  But might as well get it as right
        # as possible.

        # To announce our existence, we need:
        #   -- the timestamp we arrived at
        #   -- the manager selection atom
        #   -- the window that registered the selection
        # Of course, because Gtk is doing so much magic for us, we have to do
        # some weird tricks to get at these.

        # Ask ourselves when we acquired the selection:
        timestamp_atom = Gdk.Atom.intern("TIMESTAMP", False)
        contents = self.clipboard.wait_for_contents(timestamp_atom)
        ts_data = contents.get_data()
        log("ManagerSelection.acquire(%s) %s.wait_for_contents(%s)=%s",
            when, self.clipboard, timestamp_atom, ts_data)

        # data is a timestamp, X11 datatype is Time which is CARD32,
        # (which is 64 bits on 64-bit systems!)
        Lsize = calcsize("@L")
        if len(ts_data) == Lsize:
            ts_num = unpack("@L", ts_data[:Lsize])[0]
        else:
            ts_num = 0  # CurrentTime=0
            log.warn("invalid data for 'TIMESTAMP': %s", tuple(hex(ord(x)) for x in ts_data))
        log("selection timestamp(%s)=%s", ts_data, ts_num)
        # Calculate the X atom for this selection:
        selection_xatom = get_xatom(self.atom)
        # Ask X what window we used:
        with xsync:
            X11Window = X11WindowBindings()
            self.xid = int(X11Window.XGetSelectionOwner(self.atom))
            root_xid = X11Window.get_root_xid()
            X11Window.sendClientMessage(root_xid, root_xid, False, StructureNotifyMask,
                                        "MANAGER",
                                        ts_num, selection_xatom, self.xid)

        if old_owner != XNone and when is self.FORCE:
            # Block in a recursive mainloop until the previous owner has
            # cleared out.
            window = get_pywindow(old_owner)
            if not window:
                log(f"Previous owner {old_owner:x} is already gone? not blocking")
            else:
                log(f"got owner window {window}")
                window.set_events(window.get_events() | Gdk.EventMask.STRUCTURE_MASK)
                log("Waiting for previous owner to exit...")
                add_event_receiver(window.get_xid(), self)
                self.exit_timer = GLib.timeout_add(SELECTION_EXIT_TIMEOUT * 1000, self.exit_timeout)
                Gtk.main()
                if self.exit_timer:
                    GLib.source_remove(self.exit_timer)
                    self.exit_timer = 0
                log("...they did.")
        prop_set(self.xid, "WM_TITLE", "latin1", "Xpra_ManagerSelection%s" % self.atom)
        self.clipboard.connect("owner-change", self._owner_change)

    def exit_timeout(self) -> None:
        self.exit_timer = 0
        log.error("selection timeout")
        log.error(" the current owner did not exit")
        sys.exit(ExitCode.TIMEOUT)

    def _owner_change(self, clipboard, event) -> None:
        log(f"owner_change({clipboard}, {event}) selection={event.selection}")
        if str(event.selection) != self.atom or not event.owner:
            # log("_owner_change(..) not our selection: %s vs %s", event.selection, self.atom)
            return
        owner = event.owner.get_xid()
        log(f"owner_change({clipboard}, {event}) selection={event.selection}, owner={owner:x}, xid={self.xid:x}")
        if owner == self.xid:
            log("_owner_change(..) we still own %s", event.selection)
            return
        if self.xid:
            self.xid = 0
            self.emit("selection-lost")

    def do_x11_destroy_event(self, event) -> None:
        xid = event.window
        if xid:
            remove_event_receiver(xid, self)
        Gtk.main_quit()


GObject.type_register(ManagerSelection)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.


def main():
    from xpra.codecs.evdi.load import load_evdi_module
    load_evdi_module()
    from xpra.codecs.drm.drm import query  # @UnresolvedImport
    info = query()
    from xpra.util.str_fn import print_nested_dict
    print_nested_dict(info)
    from xpra.codecs.evdi.capture import find_evdi_devices, test_device  # @UnresolvedImport
    #from xpra.gtk.gobject_compat import register_os_signals
    #import sys
    #def handler(*args):
    #    sys.exit(0)
    #register_os_signals(handler, "evdi test")
    devices = find_evdi_devices()
    print(f"devices={devices}")
    if devices:
        for device in devices:
            test_device(device)
            break


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2013 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import stat
import sys
from subprocess import PIPE, Popen
from tempfile import NamedTemporaryFile
from typing import Any
from collections.abc import Sequence

from xpra.common import noerr
from xpra.os_util import POSIX

util_logger = None


def get_util_logger():
    global util_logger
    if not util_logger:
        from xpra.log import Logger
        util_logger = Logger("network", "util")
    return util_logger


def load_binary_file(filename) -> bytes:
    if not os.path.exists(filename):
        return b""
    try:
        with open(filename, "rb") as f:
            return f.read()
    except Exception as e:  # pragma: no cover
        log = get_util_logger()
        log.debug(f"load_binary_file({filename})", exc_info=True)
        log.warn(f"Warning: failed to load {filename!r}")
        log.warn(f" {e}")
        return b""


def filedata_nocrlf(filename: str) -> bytes:
    v = load_binary_file(filename)
    if v is None:
        log = get_util_logger()
        log.error(f"failed to load {filename!r}")
        return b""
    return v.strip(b"\n\r")


def is_socket(sockpath: str, check_uid: int = -1) -> bool:
    try:
        s = os.stat(sockpath)
    except OSError as e:
        get_util_logger().debug(f"is_socket({sockpath}) path cannot be accessed: {e}")
        # socket cannot be accessed
        return False
    if not stat.S_ISSOCK(s.st_mode):
        return False
    if check_uid >= 0:
        logger = get_util_logger()
        if s.st_uid != check_uid:
            # socket uid does not match
            logger.debug(f"is_socket({sockpath}, {check_uid}) uid {s.st_uid} does not match {check_uid}")
            return False
        logger.debug(f"is_socket({sockpath}, {check_uid}) uid matches")
    return True


def is_writable(path: str, uid: int, gid: int) -> bool:
    try:
        s = os.stat(path)
    except OSError as e:
        get_util_logger().debug(f"is_writable({path}) path cannot be accessed: {e}")
        # socket cannot be accessed
        return False
    mode = s.st_mode
    if s.st_uid == uid and mode & stat.S_IWUSR:
        # uid has write access
        return True
    if s.st_gid == gid and mode & stat.S_IWGRP:
        # gid has write access:
        return True
    return False


def stderr_print(msg: str = "") -> bool:
    stderr = sys.stderr
    if stderr:
        try:
            noerr(stderr.write, msg + "\n")
            noerr(stderr.flush)
            return True
        except (OSError, AttributeError):
            pass
    return False


def info(msg: str) -> None:
    if not stderr_print(msg) and POSIX:
        import syslog
        syslog.syslog(syslog.LOG_INFO, msg)


def warn(msg: str) -> None:
    if not stderr_print(msg) and POSIX:
        import syslog
        syslog.syslog(syslog.LOG_WARNING, msg)


def error(msg: str) -> None:
    if not stderr_print(msg) and POSIX:
        import syslog
        syslog.syslog(syslog.LOG_ERR, msg)


class CaptureStdErr:
    __slots__ = ("savedstderr", "tmp", "stderr")

    def __init__(self, *_args):
        self.savedstderr = None
        self.stderr = b""

    def __enter__(self):
        noerr(sys.stderr.flush)  # <--- important when redirecting to files
        self.savedstderr = os.dup(2)
        self.tmp = NamedTemporaryFile(prefix="stderr")
        fd = self.tmp.fileno()
        os.dup2(fd, 2)
        try:
            sys.stderr = os.fdopen(self.savedstderr, "w")
        except OSError as e:
            noerr(sys.stderr.write, f"failed to replace stderr: {e}\n")

    def __exit__(self, *_args):
        try:
            fd = self.tmp.fileno()
            os.lseek(fd, 0, 0)
            self.stderr = os.read(fd, 32768)
            self.tmp.close()
        except OSError as e:
            noerr(sys.stderr.write, f"failed to restore stderr: {e}\n")
        if self.savedstderr is not None:
            os.dup2(self.savedstderr, 2)


def path_permission_info(filename: str, ftype="") -> Sequence[str]:
    from xpra.os_util import POSIX
    if not POSIX:
        return ()
    pinfo = []
    try:
        stat_info = os.stat(filename)
        if not ftype:
            ftype = "file"
            if os.path.isdir(filename):
                ftype = "directory"
        operm = oct(stat.S_IMODE(stat_info.st_mode))
        pinfo.append(f"permissions on {ftype} {filename}: {operm}")
        # pylint: disable=import-outside-toplevel
        import pwd
        import grp
        user = pwd.getpwuid(stat_info.st_uid)[0]
        group = grp.getgrgid(stat_info.st_gid)[0]
        pinfo.append(f"ownership {user}:{group}")
    except Exception as e:
        pinfo.append(f"failed to query path information for {filename!r}: {e}")
    return tuple(pinfo)


def disable_stdout_buffering() -> None:
    import gc
    # Appending to gc.garbage is a way to stop an object from being
    # destroyed.  If the old sys.stdout is ever collected, it will
    # close() stdout, which is not good.
    gc.garbage.append(sys.stdout)
    sys.stdout = os.fdopen(sys.stdout.fileno(), 'w', 0)


def setbinarymode(fd: int) -> None:
    from xpra.os_util import WIN32
    if WIN32:
        # turn on binary mode:
        try:
            import msvcrt
            msvcrt.setmode(fd, os.O_BINARY)  # pylint: disable=no-member
        except OSError:
            get_util_logger().error("setting stdin to binary mode failed", exc_info=True)


def livefds() -> set[int]:
    live = set()
    try:
        maxfd = os.sysconf("SC_OPEN_MAX")
    except (ValueError, AttributeError):
        maxfd = 256
    for fd in range(0, maxfd):
        try:
            s = os.fstat(fd)
        except OSError:
            continue
        else:
            if s:
                live.add(fd)
    return live


def use_tty() -> bool:
    from xpra.util.env import envbool
    if envbool("XPRA_NOTTY", False):
        return False
    from xpra.platform.gui import use_stdin
    return use_stdin()


def use_gui_prompt() -> bool:
    from xpra.os_util import WIN32, OSX
    return WIN32 or OSX or not use_tty()


class umask_context:
    __slots__ = ("umask", "orig_umask")

    def __init__(self, umask):
        self.umask = umask

    def __enter__(self):
        self.orig_umask = os.umask(self.umask)

    def __exit__(self, *_args):
        os.umask(self.orig_umask)

    def __repr__(self):
        return f"umask_context({self.umask})"


def find_lib_ldconfig(libname: str) -> str:
    libname = re.escape(libname)
    arch_map = {"x86_64": "libc6,x86-64"}
    arch = arch_map.get(os.uname()[4], "libc6")
    pattern = r'^\s+lib%s\.[^\s]+ \(%s(?:,.*?)?\) => (.*lib%s[^\s]+)' % (libname, arch, libname)
    # try to find ldconfig first, which may not be on the $PATH
    # (it isn't on Debian..)
    ldconfig = "ldconfig"
    for d in ("/sbin", "/usr/sbin"):
        t = os.path.join(d, "ldconfig")
        if os.path.exists(t):
            ldconfig = t
            break
    import subprocess
    p = subprocess.Popen(f"{ldconfig} -p", stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, text=True)
    data = p.communicate()[0]
    libpath = re.search(pattern, data, re.MULTILINE)
    if libpath:
        return libpath.group(1)
    return ""


def find_lib(libname: str) -> str:
    # it would be better to rely on dlopen to find the paths,
    # but I cannot find a way of getting ctypes to tell us the path
    # it found the library in
    libpaths: list[str] = os.environ.get("LD_LIBRARY_PATH", "").split(":")
    if sys.platlibdir not in libpaths:
        libpaths.append(str(sys.platlibdir))
    for libpath in libpaths:
        if not libpath or not os.path.exists(libpath):
            continue
        libname_so = os.path.join(libpath, libname)
        if os.path.exists(libname_so):
            return libname_so
    return ""


def pollwait(process, timeout=5) -> int | None:
    from subprocess import TimeoutExpired
    try:
        return process.wait(timeout)
    except TimeoutExpired:
        return None


def find_in_PATH(command: str) -> str | None:
    path = os.environ.get("PATH", None)
    if not path:
        return None
    paths = path.split(os.pathsep)
    for p in paths:
        f = os.path.join(p, command)
        if os.path.isfile(f):
            return f
    return None


def which(command: str) -> str:
    try:
        from shutil import which
        return which(command) or ""
    except OSError:
        get_util_logger().debug(f"find_executable({command})", exc_info=True)
        return ""


def get_status_output(*args, **kwargs) -> tuple[int, Any, Any]:
    kwargs |= {
        "stdout": PIPE,
        "stderr": PIPE,
        "universal_newlines": True,
    }
    try:
        p = Popen(*args, **kwargs)
    except Exception as e:
        from xpra.log import Logger
        log = Logger("util")
        log.error(f"Error running {args},{kwargs}: {e}")
        return -1, "", ""
    stdout, stderr = p.communicate()
    return p.returncode, stdout, stderr


def get_proc_cmdline(pid: int) -> Sequence[str]:
    from xpra.os_util import POSIX
    if pid and POSIX:
        # try to find the command via /proc:
        proc_cmd_line = os.path.join("/proc", f"{pid}", "cmdline")
        if os.path.exists(proc_cmd_line):
            cmdline = load_binary_file(proc_cmd_line).rstrip(b"\0").split(b"\0")
            try:
                return tuple(x.decode() for x in cmdline)
            except UnicodeDecodeError:
                return tuple(x.decode("latin1") for x in cmdline)
    return ()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2011 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import unittest

from xpra.util.objects import AtomicInteger, MutableInteger, typedict
from xpra.util.screen import log_screen_sizes
from xpra.util.str_fn import std, alnum, nonl, pver


class TestIntegerClasses(unittest.TestCase):

    def _test_IntegerClass(self, IntegerClass):
        a = IntegerClass()
        a.increase()
        a.decrease()
        self.assertEqual(int(a), 0)
        a.increase(10)
        self.assertEqual(int(a.get()), 10)
        a.decrease()        #9
        self.assertLess(int(a), 10)
        a.decrease(19)
        self.assertGreaterEqual(int(a), -10)
        self.assertEqual(int(IntegerClass(int(str(a)))), -10)

    def test_AtomicInteger(self):
        self._test_IntegerClass(AtomicInteger)

    def test_AtomicInteger_threading(self):
        a = AtomicInteger()
        N = 5000

        def increase():
            for _ in range(N):
                a.increase()

        def decrease():
            for _ in range(N):
                a.decrease()
        T = 20
        from threading import Thread
        threads = []
        for i in range(T):
            if i % 2 == 0:
                target = increase
            else:
                target = decrease
            t = Thread(target=target, name=str(target))
            t.daemon = True
            threads.append(t)
        for t in threads:
            t.start()
        for t in threads:
            t.join()

    def test_MutableInteger(self):
        self._test_IntegerClass(MutableInteger)


class TestTypedict(unittest.TestCase):

    def test_typedict(self):
        d = typedict({
            b"bytekey" : b"bytevalue",
            "strkey" : "strvalue",
            "boolvalue" : True,
            "intpair" : (1, 2),
            "strtuple" : ["a", "b"],
        })
        #test all accessors:
        self.assertEqual(d.strget("strkey"), "strvalue")
        self.assertEqual(d.boolget("boolvalue"), True)
        self.assertEqual(d.intpair("intpair"), (1, 2))
        self.assertEqual(d.strtupleget("strtuple"), ("a", "b"))
        #now test defaults:
        self.assertEqual(d.boolget("invalidkey"), False)
        self.assertEqual(d.boolget("invalidkey", False), False)
        self.assertEqual(d.boolget("invalidkey", True), True)
        self.assertEqual(d.intget("invalidkey"), 0)
        self.assertEqual(d.intget("invalidkey", 1), 1)
        self.assertEqual(d.strget("invalidkey"), "")

    def _test_values_type(self, d, getter, value_types, values_allowed=()):
        for k in d.keys():
            v = getter(k)
            self.assertIsNotNone(v, "expected value for %s key '%s'" % (type(k), k))
            self.assertIn(type(v), value_types)
            if values_allowed:
                self.assertIn(v, values_allowed, "unexpected value for %s" % k)

    def test_strget(self):
        d = typedict({"bytekey"    : b"bytevalue",
                      "unicodekey" : "unicodevalue"})
        self._test_values_type(d, d.strget, [str, ])

    def test_intget(self):
        d = typedict({
            "strvalue": "1",
            "unicodekey": 2,
            "float-coercion": 3.14,
        })
        self._test_values_type(d, d.intget, [int], [1, 2, 3])

    def test_boolget(self):
        d = typedict({"empty-string-is-false"          : b"",
                      "False boolean stays as it is"   : False,
                      "zero is False"                  : 0})
        self._test_values_type(d, d.boolget, [bool], [False])
        d = typedict({"non-empty-string-is-true"       : "hello",
                      "True boolean stays as it is"    : True,
                      "non-zero number is True"        : -1})
        self._test_values_type(d, d.boolget, [bool], [True])

    def test_dictget(self):
        d = typedict({"nested": {}})
        self._test_values_type(d, d.dictget, [dict])

    #def intpair(self, k, default_value=None):
    #def strtupleget(self, k, default_value=[]):
    #def inttupleget(self, k, default_value=[]):
    #def listget(self, k, default_value=[], item_type=None, max_items=None):


class TestModuleFunctions(unittest.TestCase):

    def test_log_screen_sizes(self):
        #whatever you throw at it, it should continue:
        log_screen_sizes(0, 0, None)
        log_screen_sizes(None, object(), ["invalid"])

    def test_pver(self):
        self.assertEqual(pver(""), "")
        self.assertEqual(pver("any string"), "any string")
        self.assertEqual(pver((1, 2, 3)), "1.2.3")

    def test_std(self):
        self.assertEqual(std(""), "")
        self.assertEqual(std("abcd"), "abcd")
        self.assertEqual(std("r1"), "r1")
        for invalid in ("*", "\n", "\r"):
            self.assertEqual(std(invalid), "")
            self.assertEqual(std("a"+invalid+"b"), "ab")

    def test_alnum(self):
        self.assertEqual(alnum("!\"$%^&*()_+{}:@~\\<>?/*-"), "")
        self.assertEqual(alnum("aBcD123"), "aBcD123")

    def test_nonl(self):
        self.assertEqual(nonl("\n\r"), "\\n\\r")
        self.assertEqual(nonl("A\nB\rC"), "A\\nB\\rC")


def main():
    unittest.main()


if __name__ == '__main__':
    main()



# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys
import time
import os.path

from xpra.util.env import envbool
from xpra.os_util import WIN32, LINUX
from xpra.util.str_fn import memoryview_to_bytes
from xpra.log import Logger, consume_verbose_argv

log = Logger("encoder", "nvfbc")

USE_NVFBC_CUDA = envbool("XPRA_NVFBC_CUDA", False)


def get_capture_module():
    if WIN32:
        from xpra.codecs.nvidia.nvfbc import capture_win
        return capture_win
    if LINUX:
        from xpra.codecs.nvidia.nvfbc import capture_linux  # @Reimport
        return capture_linux
    return None


def get_capture_instance(cuda=USE_NVFBC_CUDA):
    fbc_module = get_capture_module()
    if not fbc_module:
        return None
    fbc_module.init_nvfbc_library()
    if cuda:
        return fbc_module.NvFBC_CUDACapture()  # @UndefinedVariable
    return fbc_module.NvFBC_SysCapture()  # @UndefinedVariable


def main(argv):
    from xpra.platform import program_context
    with program_context("NvFBC-Capture", "NvFBC Capture"):
        consume_verbose_argv(argv, "nvfbc")
        from xpra.platform.paths import get_download_dir
        from xpra.util.str_fn import print_nested_dict
        fbc_capture = get_capture_module()
        if not fbc_capture:
            raise RuntimeError("nvfbc is not supported on this platform")
        fbc_capture.init_module({})
        if WIN32:
            try:
                if "enable" in argv[1:]:
                    fbc_capture.set_enabled(True)
                    log.info("nvfbc capture enabled")
                elif "disable" in argv[1:]:
                    fbc_capture.set_enabled(False)
                    log.info("nvfbc capture disabled")
            except Exception as e:
                log(f"set_enabled for {argv} failed", exc_info=True)
                log.error("Error: cannot enable or disable NvFBC:")
                log.estr(e)
                log.error(" you may need to run this command as administrator")
                return 1
        log.info("Info:")
        print_nested_dict(fbc_capture.get_info(), print_fn=log.info)
        log.info("Status:")
        print_nested_dict(fbc_capture.get_status(), print_fn=log.info)
        try:
            log("creating test capture instance")
            c = get_capture_instance()
            log("Capture=%s", c)
            c.init_context()
            assert c.refresh()
        except Exception as e:
            log("Capture()", exc_info=True)
            log.error("Error: failed to create test capture instance:")
            log.estr(e)
            return 1
        image = c.get_image()
        assert image
        from PIL import Image
        w = image.get_width()
        h = image.get_height()
        pixels = memoryview_to_bytes(image.get_pixels())
        stride = image.get_rowstride()
        rgb_format = image.get_pixel_format()
        try:
            img = Image.frombuffer("RGB", (w, h), pixels, "raw", rgb_format, stride, 1)
            filename = os.path.join(os.path.expanduser(get_download_dir()),
                                    "screenshot-%s-%i.png" % (rgb_format, time.time()))
            img.save(filename, "png")
            log.info("screenshot saved to %s", filename)
            return 0
        except Exception as e:
            log.warn("Error: failed to save %s:", rgb_format)
            log.warn(" %s", e)
            return 1


if __name__ == "__main__":
    sys.exit(main(sys.argv))



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import math

from time import monotonic
from typing import Any, Dict, Tuple
from collections.abc import Sequence

from xpra.log import Logger

from libc.stddef cimport wchar_t, size_t
from libc.stdint cimport uintptr_t, uint8_t, uint16_t, uint32_t

log = Logger("win32", "d3d11")

ctypedef unsigned int UINT
ctypedef unsigned int HRESULT
ctypedef unsigned long ULONG
ctypedef long LONG
ctypedef unsigned int DWORD


VENDORS: Dict[int, str] = {
    0x1002: "AMD",
    0x1022: "AMD",
    0x1010: "ImgTec",
    0x10DE: "NVIDIA",
    0x13B5: "ARM",
    0x5143: "Qualcomm",
    0x163C: "Intel",
    0x8086: "Intel",
    0x8087: "Intel",
}


ctypedef struct LUID:
    DWORD LowPart
    LONG HighPart


cdef extern from "Python.h":
    int PyObject_GetBuffer(object obj, Py_buffer *view, int flags)
    void PyBuffer_Release(Py_buffer *view)
    int PyBUF_ANY_CONTIGUOUS
    object PyUnicode_FromWideChar(wchar_t *w, Py_ssize_t size)


cdef extern from "string.h":
    size_t wcslen(const wchar_t *str)


cdef extern from "guiddef.h":
    ctypedef struct GUID:
        unsigned long  Data1
        unsigned short Data2
        unsigned short Data3
        unsigned char  Data4[8]

    ctypedef GUID REFIID


cdef extern from "winerror.h":
    int DXGI_STATUS_OCCLUDED
    int DXGI_STATUS_CLIPPED
    int DXGI_STATUS_NO_REDIRECTION
    int DXGI_STATUS_NO_DESKTOP_ACCESS
    int DXGI_STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE
    int DXGI_STATUS_MODE_CHANGED
    int DXGI_STATUS_MODE_CHANGE_IN_PROGRESS
    int DXGI_ERROR_INVALID_CALL
    int DXGI_ERROR_NOT_FOUND
    int DXGI_ERROR_MORE_DATA
    int DXGI_ERROR_UNSUPPORTED
    int DXGI_ERROR_DEVICE_REMOVED
    int DXGI_ERROR_DEVICE_HUNG
    int DXGI_ERROR_DEVICE_RESET
    int DXGI_ERROR_WAS_STILL_DRAWING
    int DXGI_ERROR_FRAME_STATISTICS_DISJOINT
    int DXGI_ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE
    int DXGI_ERROR_DRIVER_INTERNAL_ERROR
    int DXGI_ERROR_NONEXCLUSIVE
    int DXGI_ERROR_NOT_CURRENTLY_AVAILABLE
    int DXGI_ERROR_REMOTE_CLIENT_DISCONNECTED
    int DXGI_ERROR_REMOTE_OUTOFMEMORY
    int D3D11_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS
    int D3D11_ERROR_FILE_NOT_FOUND
    int D3D11_ERROR_TOO_MANY_UNIQUE_VIEW_OBJECTS
    int D3D11_ERROR_DEFERRED_CONTEXT_MAP_WITHOUT_INITIAL_DISCARD
    int D3D10_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS
    int D3D10_ERROR_FILE_NOT_FOUND
    int E_NOINTERFACE
    int E_POINTER


ERRORS = {
    DXGI_STATUS_OCCLUDED: "STATUS_OCCLUDED",
    DXGI_STATUS_CLIPPED: "STATUS_CLIPPED",
    DXGI_STATUS_NO_REDIRECTION: "STATUS_NO_REDIRECTION",
    DXGI_STATUS_NO_DESKTOP_ACCESS: "STATUS_NO_DESKTOP_ACCESS",
    DXGI_STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE: "STATUS_GRAPHICS_VIDPN_SOURCE_IN_USE",
    DXGI_STATUS_MODE_CHANGED: "STATUS_MODE_CHANGED",
    DXGI_STATUS_MODE_CHANGE_IN_PROGRESS: "STATUS_MODE_CHANGE_IN_PROGRESS",
    DXGI_ERROR_INVALID_CALL: "INVALID_CALL",
    DXGI_ERROR_NOT_FOUND: "NOT_FOUND",
    DXGI_ERROR_MORE_DATA: "MORE_DATA",
    DXGI_ERROR_UNSUPPORTED: "UNSUPPORTED",
    DXGI_ERROR_DEVICE_REMOVED: "DEVICE_REMOVED",
    DXGI_ERROR_DEVICE_HUNG: "DEVICE_HUNG",
    DXGI_ERROR_DEVICE_RESET: "DEVICE_RESET",
    DXGI_ERROR_WAS_STILL_DRAWING: "WAS_STILL_DRAWING",
    DXGI_ERROR_FRAME_STATISTICS_DISJOINT: "FRAME_STATISTICS_DISJOINT",
    DXGI_ERROR_GRAPHICS_VIDPN_SOURCE_IN_USE: "GRAPHICS_VIDPN_SOURCE_IN_USE",
    DXGI_ERROR_DRIVER_INTERNAL_ERROR: "DRIVER_INTERNAL_ERROR",
    DXGI_ERROR_NONEXCLUSIVE: "NONEXCLUSIVE",
    DXGI_ERROR_NOT_CURRENTLY_AVAILABLE: "NOT_CURRENTLY_AVAILABLE",
    DXGI_ERROR_REMOTE_CLIENT_DISCONNECTED: "REMOTE_CLIENT_DISCONNECTED",
    DXGI_ERROR_REMOTE_OUTOFMEMORY: "REMOTE_OUTOFMEMORY",
    D3D11_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS: "TOO_MANY_UNIQUE_STATE_OBJECTS",
    D3D11_ERROR_FILE_NOT_FOUND: "FILE_NOT_FOUND",
    D3D11_ERROR_TOO_MANY_UNIQUE_VIEW_OBJECTS: "TOO_MANY_UNIQUE_VIEW_OBJECTS",
    D3D11_ERROR_DEFERRED_CONTEXT_MAP_WITHOUT_INITIAL_DISCARD: "DEFERRED_CONTEXT_MAP_WITHOUT_INITIAL_DISCARD",
    D3D10_ERROR_TOO_MANY_UNIQUE_STATE_OBJECTS: "TOO_MANY_UNIQUE_STATE_OBJECTS",
    D3D10_ERROR_FILE_NOT_FOUND: "FILE_NOT_FOUND",
    E_NOINTERFACE: "NOINTERFACE",
    E_POINTER: "POINTER",
}


cdef extern from "dxgi.h":
    ctypedef struct DXGI_ADAPTER_DESC:
        wchar_t  Description[128]
        UINT   VendorId
        UINT   DeviceId
        UINT   SubSysId
        UINT   Revision
        size_t DedicatedVideoMemory
        size_t DedicatedSystemMemory
        size_t SharedSystemMemory
        LUID   AdapterLuid

    # DEFINE_GUID(IID_IDXGIAdapter, 0x2411e7e1, 0x12ac, 0x4ccf, 0xbd, 0x14, 0x97, 0x9b, 0xe2, 0x52, 0x12, 0x20)
    ctypedef HRESULT (*ADAPTER_GETDESC)(IDXGIAdapter *this, DXGI_ADAPTER_DESC *desc)

    ctypedef struct IDXGIAdapterVtbl:
        ADAPTER_GETDESC GetDesc

    ctypedef struct IDXGIAdapter:
        const IDXGIAdapterVtbl *lpVtbl

    # DEFINE_GUID(IID_IDXGIDevice, 0x54ec77fa, 0x1377, 0x44e6, 0x8c, 0x32, 0x88, 0xfd, 0x5f, 0x44, 0xc8, 0x4c)
    ctypedef HRESULT (*DEVICE_GETADAPTER)(IDXGIDevice *this, IDXGIAdapter** adapter)
    ctypedef struct IDXGIDeviceVtbl:
        DEVICE_GETADAPTER GetAdapter

    ctypedef struct IDXGIDevice:
        const IDXGIDeviceVtbl *lpVtbl


cdef extern from "d3d11.h":
    ctypedef enum DXGI_FORMAT:
        DXGI_FORMAT_UNKNOWN
        DXGI_FORMAT_R32G32B32A32_TYPELESS
        DXGI_FORMAT_R32G32B32A32_FLOAT
        DXGI_FORMAT_R32G32B32A32_UINT
        DXGI_FORMAT_R32G32B32A32_SINT
        DXGI_FORMAT_R32G32B32_TYPELESS
        DXGI_FORMAT_R32G32B32_FLOAT
        DXGI_FORMAT_R32G32B32_UINT
        DXGI_FORMAT_R32G32B32_SINT
        DXGI_FORMAT_R16G16B16A16_TYPELESS
        DXGI_FORMAT_R16G16B16A16_FLOAT
        DXGI_FORMAT_R16G16B16A16_UNORM
        DXGI_FORMAT_R16G16B16A16_UINT
        DXGI_FORMAT_R16G16B16A16_SNORM
        DXGI_FORMAT_R16G16B16A16_SINT
        DXGI_FORMAT_R32G32_TYPELESS
        DXGI_FORMAT_R32G32_FLOAT
        DXGI_FORMAT_R32G32_UINT
        DXGI_FORMAT_R32G32_SINT
        DXGI_FORMAT_R32G8X24_TYPELESS
        DXGI_FORMAT_D32_FLOAT_S8X24_UINT
        DXGI_FORMAT_R32_FLOAT_X8X24_TYPELESS
        DXGI_FORMAT_X32_TYPELESS_G8X24_UINT
        DXGI_FORMAT_R10G10B10A2_TYPELESS
        DXGI_FORMAT_R10G10B10A2_UNORM
        DXGI_FORMAT_R10G10B10A2_UINT
        DXGI_FORMAT_R11G11B10_FLOAT
        DXGI_FORMAT_R8G8B8A8_TYPELESS
        DXGI_FORMAT_R8G8B8A8_UNORM
        DXGI_FORMAT_R8G8B8A8_UNORM_SRGB
        DXGI_FORMAT_R8G8B8A8_UINT
        DXGI_FORMAT_R8G8B8A8_SNORM
        DXGI_FORMAT_R8G8B8A8_SINT
        DXGI_FORMAT_R16G16_TYPELESS
        DXGI_FORMAT_R16G16_FLOAT
        DXGI_FORMAT_R16G16_UNORM
        DXGI_FORMAT_R16G16_UINT
        DXGI_FORMAT_R16G16_SNORM
        DXGI_FORMAT_R16G16_SINT
        DXGI_FORMAT_R32_TYPELESS
        DXGI_FORMAT_D32_FLOAT
        DXGI_FORMAT_R32_FLOAT
        DXGI_FORMAT_R32_UINT
        DXGI_FORMAT_R32_SINT
        DXGI_FORMAT_R24G8_TYPELESS
        DXGI_FORMAT_D24_UNORM_S8_UINT
        DXGI_FORMAT_R24_UNORM_X8_TYPELESS
        DXGI_FORMAT_X24_TYPELESS_G8_UINT
        DXGI_FORMAT_R8G8_TYPELESS
        DXGI_FORMAT_R8G8_UNORM
        DXGI_FORMAT_R8G8_UINT
        DXGI_FORMAT_R8G8_SNORM
        DXGI_FORMAT_R8G8_SINT
        DXGI_FORMAT_R16_TYPELESS
        DXGI_FORMAT_R16_FLOAT
        DXGI_FORMAT_D16_UNORM
        DXGI_FORMAT_R16_UNORM
        DXGI_FORMAT_R16_UINT
        DXGI_FORMAT_R16_SNORM
        DXGI_FORMAT_R16_SINT
        DXGI_FORMAT_R8_TYPELESS
        DXGI_FORMAT_R8_UNORM
        DXGI_FORMAT_R8_UINT
        DXGI_FORMAT_R8_SNORM
        DXGI_FORMAT_R8_SINT
        DXGI_FORMAT_A8_UNORM
        DXGI_FORMAT_R1_UNORM
        DXGI_FORMAT_R9G9B9E5_SHAREDEXP
        DXGI_FORMAT_R8G8_B8G8_UNORM
        DXGI_FORMAT_G8R8_G8B8_UNORM
        DXGI_FORMAT_BC1_TYPELESS
        DXGI_FORMAT_BC1_UNORM
        DXGI_FORMAT_BC1_UNORM_SRGB
        DXGI_FORMAT_BC2_TYPELESS
        DXGI_FORMAT_BC2_UNORM
        DXGI_FORMAT_BC2_UNORM_SRGB
        DXGI_FORMAT_BC3_TYPELESS
        DXGI_FORMAT_BC3_UNORM
        DXGI_FORMAT_BC3_UNORM_SRGB
        DXGI_FORMAT_BC4_TYPELESS
        DXGI_FORMAT_BC4_UNORM
        DXGI_FORMAT_BC4_SNORM
        DXGI_FORMAT_BC5_TYPELESS
        DXGI_FORMAT_BC5_UNORM
        DXGI_FORMAT_BC5_SNORM
        DXGI_FORMAT_B5G6R5_UNORM
        DXGI_FORMAT_B5G5R5A1_UNORM
        DXGI_FORMAT_B8G8R8A8_UNORM
        DXGI_FORMAT_B8G8R8X8_UNORM
        DXGI_FORMAT_R10G10B10_XR_BIAS_A2_UNORM
        DXGI_FORMAT_B8G8R8A8_TYPELESS
        DXGI_FORMAT_B8G8R8A8_UNORM_SRGB
        DXGI_FORMAT_B8G8R8X8_TYPELESS
        DXGI_FORMAT_B8G8R8X8_UNORM_SRGB
        DXGI_FORMAT_BC6H_TYPELESS
        DXGI_FORMAT_BC6H_UF16
        DXGI_FORMAT_BC6H_SF16
        DXGI_FORMAT_BC7_TYPELESS
        DXGI_FORMAT_BC7_UNORM
        DXGI_FORMAT_BC7_UNORM_SRGB
        DXGI_FORMAT_AYUV
        DXGI_FORMAT_Y410
        DXGI_FORMAT_Y416
        DXGI_FORMAT_NV12
        DXGI_FORMAT_P010
        DXGI_FORMAT_P016
        DXGI_FORMAT_420_OPAQUE
        DXGI_FORMAT_YUY2
        DXGI_FORMAT_Y210
        DXGI_FORMAT_Y216
        DXGI_FORMAT_NV11
        DXGI_FORMAT_AI44
        DXGI_FORMAT_IA44
        DXGI_FORMAT_P8
        DXGI_FORMAT_A8P8
        DXGI_FORMAT_B4G4R4A4_UNORM
        DXGI_FORMAT_P208
        DXGI_FORMAT_V208
        DXGI_FORMAT_V408
        DXGI_FORMAT_SAMPLER_FEEDBACK_MIN_MIP_OPAQUE
        DXGI_FORMAT_SAMPLER_FEEDBACK_MIP_REGION_USED_OPAQUE
        DXGI_FORMAT_FORCE_UINT = 0xffffffff

    ctypedef struct DXGI_SAMPLE_DESC:
        UINT Count
        UINT Quality

    ctypedef enum D3D11_USAGE:
      D3D11_USAGE_DEFAULT
      D3D11_USAGE_IMMUTABLE
      D3D11_USAGE_DYNAMIC
      D3D11_USAGE_STAGING

    ctypedef enum D3D11_CPU_ACCESS_FLAG:
        D3D11_CPU_ACCESS_WRITE
        D3D11_CPU_ACCESS_READ

    ctypedef enum D3D11_DEVICE_CONTEXT_TYPE:
        D3D11_DEVICE_CONTEXT_IMMEDIATE
        D3D11_DEVICE_CONTEXT_DEFERRED

    ctypedef enum D3D_FEATURE_LEVEL:
        # D3D_FEATURE_LEVEL_1_0_GENERIC
        # D3D_FEATURE_LEVEL_1_0_CORE
        D3D_FEATURE_LEVEL_9_1
        D3D_FEATURE_LEVEL_9_2
        D3D_FEATURE_LEVEL_9_3
        D3D_FEATURE_LEVEL_10_0
        D3D_FEATURE_LEVEL_10_1
        D3D_FEATURE_LEVEL_11_0
        D3D_FEATURE_LEVEL_11_1
        D3D_FEATURE_LEVEL_12_0
        D3D_FEATURE_LEVEL_12_1
        D3D_FEATURE_LEVEL_12_2

    ctypedef enum D3D11_FEATURE:
        D3D11_FEATURE_THREADING
        D3D11_FEATURE_DOUBLES
        D3D11_FEATURE_FORMAT_SUPPORT
        D3D11_FEATURE_FORMAT_SUPPORT2
        D3D11_FEATURE_D3D10_X_HARDWARE_OPTIONS
        D3D11_FEATURE_D3D11_OPTIONS
        D3D11_FEATURE_ARCHITECTURE_INFO
        D3D11_FEATURE_D3D9_OPTIONS
        D3D11_FEATURE_SHADER_MIN_PRECISION_SUPPORT
        D3D11_FEATURE_D3D9_SHADOW_SUPPORT
        D3D11_FEATURE_D3D11_OPTIONS1
        D3D11_FEATURE_D3D9_SIMPLE_INSTANCING_SUPPORT
        D3D11_FEATURE_MARKER_SUPPORT
        D3D11_FEATURE_D3D9_OPTIONS1
        D3D11_FEATURE_D3D11_OPTIONS2
        D3D11_FEATURE_D3D11_OPTIONS3
        D3D11_FEATURE_GPU_VIRTUAL_ADDRESS_SUPPORT
        D3D11_FEATURE_D3D11_OPTIONS4
        D3D11_FEATURE_SHADER_CACHE
        D3D11_FEATURE_D3D11_OPTIONS5
        # D3D11_FEATURE_DISPLAYABLE
        # D3D11_FEATURE_D3D11_OPTIONS6

    ctypedef enum D3D11_CREATE_DEVICE_FLAG:
        D3D11_CREATE_DEVICE_SINGLETHREADED
        D3D11_CREATE_DEVICE_DEBUG
        D3D11_CREATE_DEVICE_SWITCH_TO_REF
        D3D11_CREATE_DEVICE_PREVENT_INTERNAL_THREADING_OPTIMIZATIONS
        D3D11_CREATE_DEVICE_BGRA_SUPPORT
        D3D11_CREATE_DEVICE_DEBUGGABLE
        D3D11_CREATE_DEVICE_PREVENT_ALTERING_LAYER_SETTINGS_FROM_REGISTRY
        D3D11_CREATE_DEVICE_DISABLE_GPU_TIMEOUT
        D3D11_CREATE_DEVICE_VIDEO_SUPPORT

    ctypedef enum D3D11_VIDEO_PROCESSOR_DEVICE_CAPS:
        D3D11_VIDEO_PROCESSOR_DEVICE_CAPS_LINEAR_SPACE
        D3D11_VIDEO_PROCESSOR_DEVICE_CAPS_xvYCC
        D3D11_VIDEO_PROCESSOR_DEVICE_CAPS_RGB_RANGE_CONVERSION
        D3D11_VIDEO_PROCESSOR_DEVICE_CAPS_YCbCr_MATRIX_CONVERSION
        D3D11_VIDEO_PROCESSOR_DEVICE_CAPS_NOMINAL_RANGE

    ctypedef enum D3D11_BUS_TYPE:
        D3D11_BUS_TYPE_OTHER
        D3D11_BUS_TYPE_PCI
        D3D11_BUS_TYPE_PCIX
        D3D11_BUS_TYPE_PCIEXPRESS
        D3D11_BUS_TYPE_AGP
        D3D11_BUS_IMPL_MODIFIER_INSIDE_OF_CHIPSET
        D3D11_BUS_IMPL_MODIFIER_TRACKS_ON_MOTHER_BOARD_TO_CHIP
        D3D11_BUS_IMPL_MODIFIER_TRACKS_ON_MOTHER_BOARD_TO_SOCKET
        D3D11_BUS_IMPL_MODIFIER_DAUGHTER_BOARD_CONNECTOR
        D3D11_BUS_IMPL_MODIFIER_DAUGHTER_BOARD_CONNECTOR_INSIDE_OF_NUAE
        D3D11_BUS_IMPL_MODIFIER_NON_STANDARD

    ctypedef struct ID3D11Device:
        ID3D11DeviceVtbl *lpVtbl

    enum D3D11_RESOURCE_DIMENSION:
        D3D11_RESOURCE_DIMENSION_UNKNOWN
        D3D11_RESOURCE_DIMENSION_BUFFER
        D3D11_RESOURCE_DIMENSION_TEXTURE1D
        D3D11_RESOURCE_DIMENSION_TEXTURE2D
        D3D11_RESOURCE_DIMENSION_TEXTURE3D

    ctypedef void (*RESOURCE_GETTYPE)(ID3D11Resource *rsc, D3D11_RESOURCE_DIMENSION *dim)

    ctypedef struct ID3D11ResourceVtbl:
        RESOURCE_GETTYPE GetType

    ctypedef struct ID3D11Resource:
        const ID3D11ResourceVtbl* lpVtbl

    ctypedef struct D3D11_BOX:
        UINT left
        UINT top
        UINT front
        UINT right
        UINT bottom
        UINT back

    ctypedef struct D3D11_TEXTURE2D_DESC:
        UINT             Width
        UINT             Height
        UINT             MipLevels
        UINT             ArraySize
        DXGI_FORMAT      Format
        DXGI_SAMPLE_DESC SampleDesc
        D3D11_USAGE      Usage
        UINT             BindFlags
        UINT             CPUAccessFlags
        UINT             MiscFlags

    ctypedef ULONG (*DC_ADDREF)(ID3D11DeviceContext *context)
    ctypedef ULONG (*DC_RELEASE)(ID3D11DeviceContext *context)
    ctypedef void (*DC_BEGIN)(ID3D11DeviceContext *context)
    ctypedef void (*DC_END)(ID3D11DeviceContext *context)
    ctypedef void (*DC_FLUSH)(ID3D11DeviceContext *context)
    ctypedef UINT (*DC_GET_CONTEXT_FLAGS)(ID3D11DeviceContext *context)
    ctypedef UINT (*DC_GET_TYPE)(ID3D11DeviceContext *context)
    ctypedef void (*DC_COPYRESOURCE)(ID3D11DeviceContext *context, ID3D11Resource *dst, ID3D11Resource *src)
    ctypedef void (*DC_COPYSUBRESOURCEREGION)(ID3D11DeviceContext *context,
                                              ID3D11Resource *dst, UINT dstSubResource, UINT dstX, UINT dstY, UINT dstZ,
                                              ID3D11Resource *src, UINT srcSubResource)
    ctypedef void (*DC_UPDATESUBRESOURCEREGION)(ID3D11DeviceContext *context,
                                                ID3D11Resource  *pDstResource, UINT DstSubresource,
                                                const D3D11_BOX *pDstBox,
                                                const void      *pSrcData,
                                                UINT            SrcRowPitch,
                                                UINT            SrcDepthPitch)

    ctypedef struct ID3D11DeviceContextVtbl:
        DC_ADDREF AddRef
        DC_RELEASE Release
        DC_BEGIN Begin
        DC_END End
        DC_FLUSH Flush
        DC_GET_CONTEXT_FLAGS GetContextFlags
        DC_GET_TYPE GetType
        DC_COPYRESOURCE CopyResource
        DC_COPYSUBRESOURCEREGION CopySubresourceRegion
        DC_UPDATESUBRESOURCEREGION UpdateSubresource

    ctypedef struct ID3D11DeviceContext:
        ID3D11DeviceContextVtbl *lpVtbl

    ctypedef ULONG (*DEVICE_ADDREF)(ID3D11Device *context)
    ctypedef ULONG (*DEVICE_RELEASE)(ID3D11Device *context)
    ctypedef HRESULT (*DEVICE_QUERYINTERFACE)(ID3D11Device* pThis, REFIID *interfaceID, void** ppInterface)
    ctypedef HRESULT (*DEVICE_CREATEBUFFER)(ID3D11Device *this, void *desc, void *initialData, void **buffer)
    ctypedef HRESULT (*DEVICE_CREATETEXTURE1D)(ID3D11Device *this, void *desc, void *initialData, void **texture)
    ctypedef HRESULT (*DEVICE_CREATETEXTURE2D)(ID3D11Device *this, void *desc, void *initialData, void **texture)
    ctypedef UINT (*DEVICE_GETCREATION_FLAGS)(ID3D11Device *this)
    ctypedef int (*DEVICE_GETFEATURELEVEL)(ID3D11Device *this)
    ctypedef int (*DEVICE_GETCREATIONFLAGS)(ID3D11Device *this)
    ctypedef void (*DEVICE_GETIMMEDIATECONTEXT)(ID3D11Device *this, ID3D11DeviceContext **context)
    ctypedef HRESULT (*DEVICE_SETEXCEPTIONMODE)(ID3D11Device *this, UINT RaiseFlags)
    ctypedef UINT (*DEVICE_GETEXCEPTIONMODE)(ID3D11Device *this)

    ctypedef struct ID3D11DeviceVtbl:
        DEVICE_ADDREF AddRef
        DEVICE_RELEASE Release
        DEVICE_QUERYINTERFACE QueryInterface
        DEVICE_CREATEBUFFER CreateBuffer
        DEVICE_CREATETEXTURE1D CreateTexture1D
        DEVICE_CREATETEXTURE2D CreateTexture2D
        DEVICE_GETCREATION_FLAGS GetCreationFlags
        DEVICE_GETFEATURELEVEL GetFeatureLevel
        DEVICE_GETCREATIONFLAGS GetCreationFlags
        DEVICE_GETIMMEDIATECONTEXT GetImmediateContext
        DEVICE_SETEXCEPTIONMODE SetExceptionMode
        DEVICE_GETEXCEPTIONMODE GetExceptionMode

    ctypedef struct ID3D11Device:
        ID3D11DeviceVtbl *lpVtbl

    ctypedef enum D3D11_QUERY:
        D3D11_QUERY_EVENT
        D3D11_QUERY_OCCLUSION
        D3D11_QUERY_TIMESTAMP
        D3D11_QUERY_TIMESTAMP_DISJOINT
        D3D11_QUERY_PIPELINE_STATISTICS
        D3D11_QUERY_OCCLUSION_PREDICATE
        D3D11_QUERY_SO_STATISTICS
        D3D11_QUERY_SO_OVERFLOW_PREDICATE
        D3D11_QUERY_SO_STATISTICS_STREAM0
        D3D11_QUERY_SO_OVERFLOW_PREDICATE_STREAM0
        D3D11_QUERY_SO_STATISTICS_STREAM1
        D3D11_QUERY_SO_OVERFLOW_PREDICATE_STREAM1
        D3D11_QUERY_SO_STATISTICS_STREAM2
        D3D11_QUERY_SO_OVERFLOW_PREDICATE_STREAM2
        D3D11_QUERY_SO_STATISTICS_STREAM3
        D3D11_QUERY_SO_OVERFLOW_PREDICATE_STREAM3

    ctypedef struct D3D11_QUERY_DESC:
        D3D11_QUERY Query
        UINT MiscFlags

    # DEFINE_GUID(IID_ID3D11Query, 0xd6c00747, 0x87b7, 0x425e, 0xb8, 0x4d, 0x44, 0xd1, 0x08, 0x56, 0x0a, 0xfd)

    ctypedef void (*QUERY_GETDESC)(ID3D11Query this, D3D11_QUERY_DESC *desc)

    ctypedef struct ID3D11QueryVtbl:
        QUERY_GETDESC GetDesc

    ctypedef struct ID3D11Query:
        const ID3D11QueryVtbl *lpVtbl


FEATURE_LEVELS: Dict[int, str] = {
    # D3D_FEATURE_LEVEL_1_0_GENERIC: "1_0_GENERIC",
    # D3D_FEATURE_LEVEL_1_0_CORE: "1_0_CORE",
    D3D_FEATURE_LEVEL_9_1: (9, 1),
    D3D_FEATURE_LEVEL_9_2: (9, 2),
    D3D_FEATURE_LEVEL_9_3: (9, 3),
    D3D_FEATURE_LEVEL_10_0: (10, 0),
    D3D_FEATURE_LEVEL_10_1: (10, 1),
    D3D_FEATURE_LEVEL_11_0: (11, 0),
    D3D_FEATURE_LEVEL_11_1: (11, 1),
    D3D_FEATURE_LEVEL_12_0: (12, 0),
    D3D_FEATURE_LEVEL_12_1: (12, 1),
    D3D_FEATURE_LEVEL_12_2: (12, 2),
}

FEATURES: Dict[int, str] = {
    D3D11_FEATURE_THREADING: "THREADING",
    D3D11_FEATURE_DOUBLES: "DOUBLES",
    D3D11_FEATURE_FORMAT_SUPPORT: "FORMAT_SUPPORT",
    D3D11_FEATURE_FORMAT_SUPPORT2: "FORMAT_SUPPORT2",
    D3D11_FEATURE_D3D10_X_HARDWARE_OPTIONS: "D3D10_X_HARDWARE_OPTIONS",
    D3D11_FEATURE_D3D11_OPTIONS: "D3D11_OPTIONS",
    D3D11_FEATURE_ARCHITECTURE_INFO: "ARCHITECTURE_INFO",
    D3D11_FEATURE_D3D9_OPTIONS: "D3D9_OPTIONS",
    D3D11_FEATURE_SHADER_MIN_PRECISION_SUPPORT: "SHADER_MIN_PRECISION_SUPPORT",
    D3D11_FEATURE_D3D9_SHADOW_SUPPORT: "D3D9_SHADOW_SUPPORT",
    D3D11_FEATURE_D3D11_OPTIONS1: "D3D11_OPTIONS1",
    D3D11_FEATURE_D3D9_SIMPLE_INSTANCING_SUPPORT: "D3D9_SIMPLE_INSTANCING_SUPPORT",
    D3D11_FEATURE_MARKER_SUPPORT: "MARKER_SUPPORT",
    D3D11_FEATURE_D3D9_OPTIONS1: "D3D9_OPTIONS1",
    D3D11_FEATURE_D3D11_OPTIONS2: "D3D11_OPTIONS2",
    D3D11_FEATURE_D3D11_OPTIONS3: "D3D11_OPTIONS3",
    D3D11_FEATURE_GPU_VIRTUAL_ADDRESS_SUPPORT: "GPU_VIRTUAL_ADDRESS_SUPPORT",
    D3D11_FEATURE_D3D11_OPTIONS4: "D3D11_OPTIONS4",
    D3D11_FEATURE_SHADER_CACHE: "SHADER_CACHE",
    D3D11_FEATURE_D3D11_OPTIONS5: "D3D11_OPTIONS5",
    # D3D11_FEATURE_DISPLAYABLE: "DISPLAYABLE",
    # D3D11_FEATURE_D3D11_OPTIONS6: "D3D11_OPTIONS6",
}

DEVICE_FLAGS: Dict[int, str] = {
    D3D11_CREATE_DEVICE_SINGLETHREADED: "SINGLETHREADED",
    D3D11_CREATE_DEVICE_DEBUG: "DEBUG",
    D3D11_CREATE_DEVICE_SWITCH_TO_REF: "SWITCH_TO_REF",
    D3D11_CREATE_DEVICE_PREVENT_INTERNAL_THREADING_OPTIMIZATIONS: "PREVENT_INTERNAL_THREADING_OPTIMIZATIONS",
    D3D11_CREATE_DEVICE_BGRA_SUPPORT: "BGRA_SUPPORT",
    D3D11_CREATE_DEVICE_DEBUGGABLE: "DEBUGGABLE",
    D3D11_CREATE_DEVICE_PREVENT_ALTERING_LAYER_SETTINGS_FROM_REGISTRY: "PREVENT_ALTERING_LAYER_SETTINGS_FROM_REGISTRY",
    D3D11_CREATE_DEVICE_DISABLE_GPU_TIMEOUT: "DISABLE_GPU_TIMEOUT",
    D3D11_CREATE_DEVICE_VIDEO_SUPPORT: "VIDEO_SUPPORT",
}


cdef class D3D11DeviceContext:
    cdef ID3D11Device *device
    cdef ID3D11DeviceContext *context

    def __init__(self, devptr: uintptr_t):
        assert devptr, "no device"
        self.device = <ID3D11Device *> devptr
        self.context = NULL

    def __enter__(self):
        self.device.lpVtbl.GetImmediateContext(self.device, &self.context)
        assert self.context
        return self

    def __exit__(self, *_args):
        assert self.context
        self.context.lpVtbl.Release(self.context)
        self.context = NULL

    def __repr__(self):
        return "ImmediateContext(%s)" % self.d3d11device

    def flush(self) -> None:
        assert self.context
        self.context.lpVtbl.Flush(self.context)

    def copy_resource(self, uintptr_t dst, uintptr_t src) -> None:
        log("D3D11DeviceContext.copy_resource(%#x, %#x)", dst, src)
        assert self.context
        assert dst and src
        cdef ID3D11Resource *r_dst = <ID3D11Resource*> dst
        cdef ID3D11Resource *r_src = <ID3D11Resource*> src
        self.context.lpVtbl.CopyResource(self.context, r_dst, r_src)

    def get_info(self) -> Dict[str, Any]:
        assert self.context
        flags = self.context.lpVtbl.GetContextFlags(self.context)
        immediate = bool(self.context.lpVtbl.GetType(self.context) & D3D11_DEVICE_CONTEXT_IMMEDIATE)
        flags_strs = tuple(name for value, name in DEVICE_FLAGS.items() if flags & value)
        return {
            "address": <uintptr_t> self.context,
            "flags": flags_strs,
            "immediate": immediate,
        }


cdef class D3D11Device:
    cdef ID3D11Device *device

    def __init__(self, devptr: uintptr_t):
        assert devptr, "no device"
        self.device = <ID3D11Device *> devptr

    def get_device_context(self):
        return D3D11DeviceContext(<uintptr_t> self.device)

    def Release(self):
        self.device.lpVtbl.Release(self.device)

    def get_immediate_context(self) -> D3D11DeviceContext:
        return D3D11DeviceContext(self)

    def __repr__(self):
        return "D3D11Device(%#x)" % <uintptr_t> self.device

    def get_info(self) -> Dict[str, Any]:
        level = self.device.lpVtbl.GetFeatureLevel(self.device)
        emode = self.device.lpVtbl.GetExceptionMode(self.device)
        # D3D11_CREATE_DEVICE_SINGLETHREADED
        info = {
            "address": <uintptr_t> self.device,
            "feature-level": FEATURE_LEVELS.get(level, level),
            "exception-mode": emode,
        }
        # get a IDXGIDevice interface from this device:
        cdef GUID device_guid               # IID_IDXGIDevice
        set_GUID(&device_guid, 0x54ec77fa, 0x1377, 0x44e6, 0x8c, 0x32, 0x88, 0xfd, 0x5f, 0x44, 0xc8, 0x4c)
        cdef IDXGIDevice *dxgi_device = NULL
        cdef HRESULT res = self.device.lpVtbl.QueryInterface(self.device, &device_guid, <void**> &dxgi_device)
        log("D3D11Device.QueryInterface(DXGIDevice)=%i, dxgi_device=%#x", res, <uintptr_t> dxgi_device)
        if res:
            log.warn("Warning: failed to get DXGIDevice interface: %s", ERRORS.get(res, res))
            return info
        assert dxgi_device
        # get the adapter interface:
        cdef GUID adapter_guid     # IID_IDXGIAdapter
        set_GUID(&adapter_guid, 0x2411e7e1, 0x12ac, 0x4ccf, 0xbd, 0x14, 0x97, 0x9b, 0xe2, 0x52, 0x12, 0x20)
        cdef IDXGIAdapter *adapter
        res = dxgi_device.lpVtbl.GetAdapter(dxgi_device, &adapter)
        if res:
            log.warn("Warning: failed to get IDXGIAdapter: %s", ERRORS.get(res, res))
            return info
        assert adapter
        # get the adapter description:
        cdef DXGI_ADAPTER_DESC desc
        res = adapter.lpVtbl.GetDesc(adapter, &desc)
        if res:
            log.warn("Warning: failed to get adapter description: %s", ERRORS.get(res, res))
            return info

        cdef size_t size = wcslen(desc.Description)
        description = PyUnicode_FromWideChar(desc.Description, size)
        info.update({
            "description": description,
            "vendor": VENDORS.get(desc.VendorId, desc.VendorId),
            "device": desc.DeviceId,
            "subsystem": desc.SubSysId,
            "revision": desc.Revision,
            "memory": {
                "video": desc.DedicatedVideoMemory,
                "system": desc.DedicatedSystemMemory,
                "shared": desc.SharedSystemMemory,
            },
            "luid": "%#x:%#x" % (desc.AdapterLuid.HighPart, desc.AdapterLuid.LowPart),
        })
        return info


cdef void set_GUID(GUID* guid, uint32_t data1, uint16_t data2, uint16_t data3,
                   uint8_t data40, uint8_t data41, uint8_t data42, uint8_t data43,
                   uint8_t data44, uint8_t data45, uint8_t data46, uint8_t data47):
    guid.Data1 = data1
    guid.Data2 = data2
    guid.Data3 = data3
    guid.Data4[0] = data40
    guid.Data4[1] = data41
    guid.Data4[2] = data42
    guid.Data4[3] = data43
    guid.Data4[4] = data44
    guid.Data4[5] = data45
    guid.Data4[6] = data46
    guid.Data4[7] = data47



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2015 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Final

import dbus.service  # @UnresolvedImport
from dbus import PROPERTIES_IFACE  # @UnresolvedImport
from dbus.exceptions import DBusException  # @UnresolvedImport

from xpra.dbus.helper import dbus_to_native
from xpra.dbus.common import init_session_bus
from xpra.util.objects import AtomicInteger
from xpra.common import ConnectionMessage
from xpra.log import Logger

log = Logger("dbus", "server")

BUS_NAME: Final[str] = "org.xpra.Server"
INTERFACE: Final[str] = "org.xpra.Client"
PATH: Final[str] = "/org/xpra/Client"


def n(*args):
    return dbus_to_native(*args)


def ni(*args):
    return int(n(*args))


def nb(*args):
    return bool(n(*args))


def ns(*args):
    return str(n(*args))


sequence = AtomicInteger()


class DBUS_Source(dbus.service.Object):
    SUPPORTS_MULTIPLE_OBJECT_PATHS = True

    def __init__(self, source=None, extra=""):
        self.source = source
        session_bus = init_session_bus()
        name = BUS_NAME
        self.path = PATH + str(sequence.increase())
        if extra:
            name += extra.replace(".", "_").replace(":", "_")
        bus_name = dbus.service.BusName(name, session_bus)
        super().__init__(bus_name, self.path)
        self.log("(%s)", source)
        self._properties = {
            "bell": ("send_bell", ni),
            "cursors": ("send_cursors", ni),
            "notifications": ("send_notifications", ni),
        }

    def __str__(self):
        return f"DBUS_Source({BUS_NAME}:{self.path})"

    def cleanup(self):
        try:
            log("calling %s", self.remove_from_connection)
            self.remove_from_connection()
        except Exception as e:
            log.error("Error removing the source's DBUS server:")
            log.estr(e)

    def log(self, fmt, *args):
        log("%s" + fmt, INTERFACE, *args)

    @dbus.service.method(PROPERTIES_IFACE, in_signature='s', out_signature='v')
    def Get(self, property_name):
        conv = self._properties.get(property_name)
        if conv is None:
            raise DBusException("invalid property")
        server_property_name, _ = conv
        v = getattr(self.source, server_property_name)
        self.log(".Get(%s)=%s", property_name, v)
        return v

    @dbus.service.method(PROPERTIES_IFACE, in_signature='', out_signature='a{sv}')
    def GetAll(self, interface_name):
        if interface_name == INTERFACE:
            v = {x: self.Get(x) for x in self._properties.keys()}
        else:
            v = {}
        self.log(".GetAll(%s)=%s", interface_name, v)
        return v

    @dbus.service.method(PROPERTIES_IFACE, in_signature='ssv')
    def Set(self, interface_name, property_name, new_value):
        self.log(".Set(%s, %s, %s)", interface_name, property_name, new_value)
        conv = self._properties.get(property_name)
        if conv is None:
            raise DBusException("invalid property")
        server_property_name, validator = conv
        assert hasattr(self.source, server_property_name)
        setattr(self.source, server_property_name, validator(new_value))

    @dbus.service.signal(PROPERTIES_IFACE, signature='sa{sv}as')
    def PropertiesChanged(self, interface_name, changed_properties, invalidated_properties):
        self.log(f"PropertiesChanged({interface_name}, {n(changed_properties)}, {n(invalidated_properties)})")

    @dbus.service.method(INTERFACE, in_signature='b')
    def ShowDesktop(self, show):
        show = nb(show)
        self.log(".ShowDesktop(%s)", show)
        self.source.show_desktop(show)

    @dbus.service.method(INTERFACE, in_signature='i')
    def RaiseWindow(self, wid):
        wid = ni(wid)
        self.log(".RaiseWindow(%s)", wid)
        self.source.raise_window(wid)

    @dbus.service.method(INTERFACE, in_signature='')
    def ResetWindowFilters(self):
        self.log(".ResetWindowFilters()")
        self.source.reset_window_filters()

    @dbus.service.method(INTERFACE, in_signature='sssv')
    def AddWindowFilter(self, object_name, property_name, operator, value):
        self.log(".AddWindowFilter%s", (object_name, property_name, operator, value))
        self.source.add_window_filter(ns(object_name), ns(property_name), ns(operator), n(value))

    @dbus.service.method(INTERFACE, out_signature='as')
    def GetAllWindowFilters(self):
        v = [str(x) for x in self.source.get_all_window_filters()]
        self.log(".GetAllWindowFilters()=%s", v)
        return v

    @dbus.service.method(INTERFACE, in_signature='')
    def SetDefaultKeymap(self):
        self.log(".SetDefaultKeymap()")
        self.source.set_default_keymap()

    @dbus.service.method(INTERFACE, in_signature='')
    def Suspend(self):
        self.log(".Suspend()")
        self.source.go_idle()

    @dbus.service.method(INTERFACE, in_signature='')
    def Resume(self):
        self.log(".Resume()")
        self.source.no_idle()

    @dbus.service.method(INTERFACE, in_signature='s')
    def StartSpeaker(self, codec):
        codec = ns(codec)
        self.log(".StartSpeaker(%s)", codec)
        self.source.start_sending_audio(codec)

    @dbus.service.method(INTERFACE, in_signature='')
    def StopSpeaker(self):
        self.log(".StopSpeaker()")
        self.source.stop_sending_audio()

    @dbus.service.method(INTERFACE, in_signature='i')
    def SetAVSyncDelay(self, delay):
        d = ni(delay)
        self.log(".SetAVSyncDelay(%i)", d)
        self.source.set_av_sync_delay(d)

    @dbus.service.method(INTERFACE, in_signature='as')
    def SendClientCommand(self, args):
        cmd = n(args)
        self.log(".SendClientCommand(%s)", cmd)
        self.source.send_client_command(*cmd)

    @dbus.service.method(INTERFACE, in_signature='s')
    def Detach(self, reason):
        proto = self.source.proto
        rs = ns(reason)
        self.log(".SendClientCommand(%s) protocol=%s", rs, proto)
        assert proto, "no connection"
        proto.send_disconnect(ConnectionMessage.DETACH_REQUEST, rs)



# This file is part of Xpra.
# Copyright (C) 2024 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys

from xpra.exit_codes import ExitCode


def main(args) -> ExitCode:
    from xpra.platform import program_context
    with program_context("Wait-for-Wayland", "Wait for Wayland"):
        from xpra.wayland.wait_for_display import wait_for_wayland_display
        if args:
            display = args[0]
        else:
            display = os.environ.get("WAYLAND_DISPLAY", "")
        try:
            wait_for_wayland_display(display)
            return ExitCode.OK
        except RuntimeError as e:
            sys.stderr.write(f"{e}\n")
            sys.stderr.flush()
            return ExitCode.FAILURE


if __name__ == "__main__":  # pragma: no cover
    main(sys.argv[1:])



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from time import time
from email.utils import formatdate

from xpra.util.env import envint
from xpra.util.str_fn import strtobytes

SERVER_NAME = "xpra/aioquic"
USER_AGENT = "xpra/aioquic"

MAX_DATAGRAM_FRAME_SIZE = envint("XPRA_MAX_DATAGRAM_FRAME_SIZE", 65536)


def http_date() -> str:
    """ GMT date in a format suitable for http headers """
    return formatdate(time(), usegmt=True)


def binary_headers(headers: dict) -> list[tuple[bytes, bytes]]:
    """ aioquic expects the headers as a list of binary pairs """
    return [(strtobytes(k), strtobytes(v)) for k, v in headers.items()]


def override_aioquic_logger() -> None:
    from xpra.log import Logger, is_debug_enabled
    import logging
    logger = logging.getLogger("quic")
    logger.propagate = False
    logger.handlers.clear()
    # warning: don't use 'quic' as first argument to Logger
    # as this would create a logging loop
    aioquic_logger = Logger("network", "quic", "verbose")
    assert aioquic_logger._logger != logger
    if is_debug_enabled("quic"):
        aioquic_logger.setLevel(logging.DEBUG)
    else:
        aioquic_logger.setLevel(logging.WARN)
    logger.addHandler(aioquic_logger)  # type: ignore



# This file is part of Xpra.
# Copyright (C) 2022 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

#cython: wraparound=False

import select
import binascii
from time import monotonic
from typing import Tuple, List
from collections.abc import Sequence

from xpra.util.env import envbool
from xpra.util.str_fn import bytestostr, strtobytes, memoryview_to_bytes

from xpra.log import Logger
log = Logger("evdi")

from libc.string cimport memset, memcpy
from xpra.buffers.membuf cimport getbuf, MemBuf
from libc.stdint cimport uintptr_t, uint8_t, uint16_t, int32_t, uint32_t

DEF DRM_MODE_DPMS_ON = 0
DEF DRM_MODE_DPMS_STANDBY = 1
DEF DRM_MODE_DPMS_SUSPEND = 2
DEF DRM_MODE_DPMS_OFF = 3

SAVE_TO_FILE = envbool("XPRA_SAVE_TO_FILE")

#https://github.com/linuxhw/EDID/tree/master/
#EDIDv2_1280x720:
#edid_hex = b"00ffffffffffff004e845d00010000000115010380311c782a0dc9a05747982712484c20000001010101010101010101010101010101011d007251d01e2046285500e812110000188c0ad08a20e02d10103e9600e81211000018000000fc0048444d492054560a2020202020000000fd00313d0f2e08000a202020202020018e02031d714701020384111213230907078301000068030c001000b82d00011d007251d01e206e285500e8121100001e000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000039"
#edid_hex = b"00ffffffffffff0005b4380001010101020d0103801e17782a80f8a3554799240d4d50bfee00614c310a0101010101010101010181c064190040410026301888360030e410000018000000fd00324b1e3e08000a202020202020000000ff003233363233303230303037340a000000fc004c322d313530542b202020200a00ef"
#800x600:
DEFAULT_EDID = binascii.unhexlify(b"00ffffffffffff0031d8000000000000051601036d1b1478ea5ec0a4594a982520505401000045400101010101010101010101010101a00f200031581c202880140015d01000001e000000ff004c696e75782023300a20202020000000fd003b3d242605000a202020202020000000fc004c696e757820535647410a202000c2")
DEFAULT_SIZE = 800*600


cdef extern from "evdi_lib.h":
    ctypedef struct evdi_lib_version:
        int version_major
        int version_minor
        int version_patchlevel

    void evdi_get_lib_version(evdi_lib_version *version)

    ctypedef struct evdi_device_context:
        pass
    ctypedef evdi_device_context *evdi_handle
    ctypedef int evdi_selectable

    ctypedef enum evdi_device_status:
        AVAILABLE
        UNRECOGNIZED
        NOT_PRESENT

    ctypedef struct evdi_rect:
        int x1, y1, x2, y2

    ctypedef struct evdi_mode:
        int width
        int height
        int refresh_rate
        int bits_per_pixel
        unsigned int pixel_format

    ctypedef struct evdi_buffer:
        int id
        void *buffer
        int width
        int height
        int stride
        evdi_rect *rects
        int rect_count

    ctypedef struct evdi_cursor_set:
        int32_t hot_x
        int32_t hot_y
        uint32_t width
        uint32_t height
        uint8_t enabled
        uint32_t buffer_length
        uint32_t *buffer
        uint32_t pixel_format
        uint32_t stride

    ctypedef struct evdi_cursor_move:
        int32_t x
        int32_t y

    ctypedef struct evdi_ddcci_data:
        uint16_t address
        uint16_t flags
        uint32_t buffer_length
        uint8_t *buffer

    ctypedef struct evdi_event_context:
        void (*dpms_handler)(int dpms_mode, void *user_data) noexcept
        void (*mode_changed_handler)(evdi_mode mode, void *user_data) noexcept
        void (*update_ready_handler)(int buffer_to_be_updated, void *user_data) noexcept
        void (*crtc_state_handler)(int state, void *user_data) noexcept
        void (*cursor_set_handler)(evdi_cursor_set cursor_set, void *user_data) noexcept
        void (*cursor_move_handler)(evdi_cursor_move cursor_move, void *user_data) noexcept
        void (*ddcci_data_handler)(evdi_ddcci_data ddcci_data, void *user_data) noexcept
        void *user_data

    ctypedef struct evdi_logging:
        void (*function)(void *user_data, const char *fmt, ...) noexcept
        void *user_data;

    #define EVDI_INVALID_HANDLE NULL
    evdi_device_status evdi_check_device(int device)
    evdi_handle evdi_open(int device)
    int evdi_add_device()
    evdi_handle evdi_open_attached_to_fixed(const char *sysfs_parent_device, size_t length)

    void evdi_close(evdi_handle handle)
    void evdi_disconnect(evdi_handle handle)

    void evdi_grab_pixels(evdi_handle handle, evdi_rect *rects, int *num_rects)
    void evdi_register_buffer(evdi_handle handle, evdi_buffer buffer)
    void evdi_unregister_buffer(evdi_handle handle, int bufferId)
    bint evdi_request_update(evdi_handle handle, int bufferId)
    void evdi_ddcci_response(evdi_handle handle, const unsigned char *buffer,
                             const uint32_t buffer_length, const bint result)

    void evdi_handle_events(evdi_handle handle, evdi_event_context *evtctx)
    evdi_selectable evdi_get_event_ready(evdi_handle handle)
    void evdi_set_logging(evdi_logging evdi_logging)

    void evdi_connect2(evdi_handle handle, const unsigned char *edid,
          const unsigned int edid_length,
          const uint32_t pixel_area_limit,
          const uint32_t pixel_per_second_limit)
    void evdi_enable_cursor_events(evdi_handle handle, bint enable)


STATUS_STR = {
    AVAILABLE       : "available",
    UNRECOGNIZED    : "unrecognized",
    NOT_PRESENT     : "not-present",
}

MODE_STR = {
    DRM_MODE_DPMS_ON        : "ON",
    DRM_MODE_DPMS_STANDBY   : "STANDBY",
    DRM_MODE_DPMS_SUSPEND   : "SUSPEND",
    DRM_MODE_DPMS_OFF       : "OFF",
}


def get_version() -> Sequence[int]:
    cdef evdi_lib_version version
    evdi_get_lib_version(&version)
    return (version.version_major, version.version_minor, version.version_patchlevel)


cdef void evdi_logging_function(void *user_data, const char *fmt, ...) noexcept:
    pybytes = fmt[:]
    log(f"evdi: %s", pybytes.decode("latin1"))


def capture_logging() -> None:
    cdef evdi_logging log_config
    log_config.function = &evdi_logging_function
    log_config.user_data = NULL
    evdi_set_logging(log_config)


def reset_logging() -> None:
    cdef evdi_logging log_config
    log_config.function = NULL
    log_config.user_data = NULL
    evdi_set_logging(log_config)


#maps device numbers to our device object:
devices = {}

cdef void dpms_handler(int dpms_mode, void *user_data) noexcept:
    log(f"dpms_handler({dpms_mode}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.dpms_handler(dpms_mode)

cdef void mode_changed_handler(evdi_mode mode, void *user_data) noexcept:
    log(f"mode_changed_handler({mode.width}x{mode.height}-{mode.bits_per_pixel}@{mode.refresh_rate}, %#x)",
        <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.mode_changed_handler(mode)

cdef void update_ready_handler(int buffer_to_be_updated, void *user_data) noexcept:
    log(f"update_ready_handler({buffer_to_be_updated}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.update_ready_handler(buffer_to_be_updated)

cdef void crtc_state_handler(int state, void *user_data) noexcept:
    log(f"crtc_state_handler({state}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.crtc_state_handler(state)

cdef void cursor_set_handler(evdi_cursor_set cursor_set, void *user_data) noexcept:
    log(f"cursor_set_handler({cursor_set.width}x{cursor_set.height}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.cursor_set_handler(cursor_set)

cdef void cursor_move_handler(evdi_cursor_move cursor_move, void *user_data) noexcept:
    log(f"cursor_move_handler({cursor_move.x}x{cursor_move.y}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.cursor_move_handler(cursor_move)

cdef void ddcci_data_handler(evdi_ddcci_data ddcci_data, void *user_data) noexcept:
    log(f"ddcci_data_handler({ddcci_data.address:x}, %#x)", <uintptr_t> user_data)
    cdef EvdiDevice evdi_device = devices.get(<uintptr_t> user_data)
    if evdi_device:
        evdi_device.ddcci_data_handler(ddcci_data)


cdef class EvdiDevice:
    cdef int device
    cdef object damage_cb
    cdef evdi_handle handle
    cdef evdi_event_context event_context
    cdef object buffers
    cdef evdi_mode mode
    cdef int dpms_mode
    cdef object edid
    cdef int export_buffer

    def __init__(self, int device, damage_cb=None):
        self.device = device
        self.damage_cb = damage_cb
        self.handle = NULL
        self.buffers = {}
        self.edid = None
        self.dpms_mode = DRM_MODE_DPMS_OFF
        self.export_buffer = 0
        memset(&self.mode, 0, sizeof(evdi_mode))
        memset(&self.event_context, 0, sizeof(evdi_event_context))
        self.event_context.dpms_handler = &dpms_handler
        self.event_context.mode_changed_handler = &mode_changed_handler
        self.event_context.update_ready_handler = &update_ready_handler
        self.event_context.crtc_state_handler = &crtc_state_handler
        self.event_context.cursor_set_handler = &cursor_set_handler
        self.event_context.cursor_move_handler = &cursor_move_handler
        self.event_context.ddcci_data_handler = &ddcci_data_handler
        self.event_context.user_data = <void *> (<uintptr_t> device)
        devices[device] = self

    def __repr__(self):
        return f"EvdiDevice({self.device} - {self.mode.width}x{self.mode.height})"

    def open(self) -> None:
        if self.handle:
            raise RuntimeError("this evdi device is already open")
        if self.device >=0:
            self.handle = evdi_open(self.device)
        else:
            self.handle = evdi_open_attached_to_fixed(NULL, 0)
        if not self.handle:
            raise ValueError(f"cannot open {self.device}")

    def close(self) -> None:
        h = self.handle
        if h:
            self.handle = NULL
            evdi_close(h)


    cdef void dpms_handler(self, int dpms_mode) noexcept:
        log(f"dpms_handler({dpms_mode}) %s", MODE_STR.get(dpms_mode, "INVALID"))
        if self.dpms_mode==dpms_mode:
            #unchanged
            return
        self.dpms_mode = dpms_mode
        if dpms_mode!=DRM_MODE_DPMS_ON:
            self.unregister_buffers()

    cdef void mode_changed_handler(self, evdi_mode mode):
        log(f"mode_changed_handler({mode.width}x{mode.height}-{mode.bits_per_pixel}@{mode.refresh_rate}")
        memcpy(&self.mode, &mode, sizeof(evdi_mode))
        self.unregister_buffers()
        log(f"mode_changed_handler dpms_mode={MODE_STR.get(self.dpms_mode, self.dpms_mode)}")
        if self.dpms_mode==DRM_MODE_DPMS_ON and self.mode.width>0 and self.mode.height>0:
            for buf_id in (1, 2):
                if buf_id not in self.buffers:
                    self.register_buffer(buf_id)
            buf_id = 1
            self.export_buffer = buf_id
            log(f"using buffer {buf_id}")
            if self.request_update(buf_id):
                self.grab_pixels(buf_id)

    cdef void update_ready_handler(self, int buffer_to_be_updated):
        log(f"update_ready_handler({buffer_to_be_updated})")
        self.grab_pixels(buffer_to_be_updated)

    def grab_pixels(self, buf_id) -> Tuple:
        if not self.handle:
            raise RuntimeError("no device handle")
        buf = self.buffers.get(buf_id)
        if not buf:
            raise ValueError(f"unknown buffer {buf_id}")
        cdef int nrects = 128
        cdef evdi_rect[128] rects
        evdi_grab_pixels(self.handle, rects, &nrects)
        cdef int rowstride = self.mode.width*4
        if SAVE_TO_FILE:
            try:
                from PIL import Image
                from PIL import __version__ as pil_version
                # older versions of PIL cannot use the memoryview directly:
                try:
                    major = int(pil_version.split(".")[0])
                except ValueError:
                    major = 0
                if major < 10:
                    pixels = memoryview(buf)
                else:
                    pixels = memoryview_to_bytes(memoryview(buf))
                pil_image = Image.frombuffer("RGBA", (self.mode.width, self.mode.height), pixels, "raw", "BGRA", rowstride)
                pil_image = pil_image.convert("RGB")
                filename = f"{monotonic()}.jpg"
                pil_image.save(filename, "JPEG")
                log(f"saved to {filename}")
                if nrects:
                    for i in range(nrects):
                        log(" %i : %i,%i to %i,%i", i, rects[i].x1, rects[i].y1, rects[i].x2, rects[i].y2)
                        w = rects[i].x2 - rects[i].x1
                        h = rects[i].y2 - rects[i].y1
                        if w>0 and h>0 and (w<self.mode.width or h<self.mode.height):
                            sub = pil_image.crop((rects[i].x1, rects[i].y1, rects[i].x2, rects[i].y2))
                            sub.save(f"{monotonic()}-{i}.jpg", "JPEG")
            except KeyboardInterrupt as e:
                log(f"{e}")
                self.cleanup()
                return
            except Exception:
                self.cleanup()
                return
        areas = tuple((rects[i].x1, rects[i].y1, rects[i].x2 - rects[i].x1, rects[i].y2 - rects[i].y1) for i in range(nrects))
        log(f"evdi_grab_pixels(%#x, %#x, {nrects}) areas={areas}", <uintptr_t> self.handle, <uintptr_t> rects)
        cdef int buf_size = rowstride * self.mode.height
        buf_slice = memoryview(buf)[:buf_size]
        damage_cb = self.damage_cb
        if damage_cb:
            self.damage_cb(self.mode.width, self.mode.height, buf_slice, areas)
        #toggle buffer:
        self.export_buffer = 3-self.export_buffer
        return self.mode.width, self.mode.height, buf_slice, areas


    cdef void crtc_state_handler(self, int state):
        log(f"crtc_state_handler({state})")

    cdef void cursor_set_handler(self, evdi_cursor_set cursor_set):
        log(f"cursor_set_handler({cursor_set.width}x{cursor_set.height})")

    cdef void cursor_move_handler(self, evdi_cursor_move cursor_move):
        log(f"cursor_move_handler({cursor_move.x}x{cursor_move.y})")

    cdef void ddcci_data_handler(self, evdi_ddcci_data ddcci_data):
        log("ddcci_data_handler(%#x)", ddcci_data.address)

    def enable_cursor_events(self, enable=True) -> None:
        evdi_enable_cursor_events(self.handle, int(enable))


    def connect(self, edid=DEFAULT_EDID) -> None:
        if not self.handle:
            raise RuntimeError("no device handle")
        if self.edid:
            raise RuntimeError("device is already connected")
        self.edid = strtobytes(edid)
        cdef uint32_t pixel_area_limit = 1920*1080
        if edid==DEFAULT_EDID:
            pixel_area_limit = DEFAULT_SIZE
        cdef int Hz = 60
        try:
            from pyedid import parse_edid
            edid_data = parse_edid(self.edid)._asdict()
            log.error(f"edid: f{edid_data}")
        except ImportError as e:
            log.warn("Warning: cannot parse EDID")
            log.warn(f" {e}")
        except ValueError as e:
            log.warn("Warning: invalid EDID data")
            log.warn(f" {e}")
        else:
            log.info(f"evdi using monitor edid: {edid_data}")
            pixel_area_limit = DEFAULT_SIZE
            maxw = 800
            maxh = 600
            for w, h, hz in edid_data.get("resolutions", ()):
                maxw = max(w, maxw)
                maxh = max(h, maxh)
                Hz = max(Hz, round(hz))
            pixel_area_limit = maxw*maxh
        cdef const unsigned char *edid_bin = self.edid
        cdef unsigned int edid_length = len(self.edid)
        cdef uint32_t pixel_per_second_limit = pixel_area_limit*Hz
        log(f"connect with edid {edid!r} (length={edid_length})")
        evdi_connect2(self.handle, edid_bin, <const unsigned int> edid_length,
                     <const uint32_t> pixel_area_limit,
                     <const uint32_t> pixel_per_second_limit)

    def disconnect(self):
        e = self.edid
        if e:
            self.edid = None
            evdi_disconnect(self.handle)

    def get_event_fd(self) -> int:
        return evdi_get_event_ready(self.handle)

    def handle_events(self) -> None:
        if not self.handle:
            raise RuntimeError("no device handle")
        if not self.edid:
            raise RuntimeError("device is not connected")
        evdi_handle_events(self.handle, &self.event_context)

    def handle_all_events(self) -> None:
        if not self.handle:
            raise RuntimeError("no device handle")
        if not self.edid:
            raise RuntimeError("device is not connected")
        cdef evdi_selectable fd = evdi_get_event_ready(self.handle)
        log(f"handle_all_events() fd={fd}")
        while self.handle!=NULL and self.edid:
            r = select.select([fd], [], [], 0)
            log(f"handle_all_events() select(..)={r}")
            if fd not in r[0]:
                break
            evdi_handle_events(self.handle, &self.event_context)

    def event_loop(self, run_time=10) -> None:
        cdef evdi_selectable fd = evdi_get_event_ready(self.handle)
        log(f"handle_events() fd={fd}")
        start = monotonic()
        while monotonic()-start<run_time and self.handle!=NULL and self.edid:
            log("waiting for events")
            r = select.select([fd], [], [], 0.020)
            log(f"select(..)={r}")
            if self.handle==NULL or not self.edid:
                break
            if fd in r[0]:
                evdi_handle_events(self.handle, &self.event_context)
            if self.buffers and self.export_buffer:
                self.refresh()

    def refresh(self) -> Tuple | None:
        buf_id = self.export_buffer
        #log(f"refresh() export_buffer={buf_id}")
        if buf_id in (1, 2):
            r = self.request_update(buf_id)
            if r:
                return self.grab_pixels(buf_id)
        return None

    def register_buffer(self, int buf_id) -> None:
        cdef evdi_buffer buf
        buf.id = buf_id
        buf.width = self.mode.width
        buf.height = self.mode.height
        buf.stride = buf.width*self.mode.bits_per_pixel//8
        buf.rect_count = 0
        buf.rects = NULL
        cdef MemBuf pybuf = getbuf(self.mode.width*self.mode.height*4)
        buf.buffer = <void *>pybuf.get_mem()
        evdi_register_buffer(self.handle, buf)
        log(f"register_buffer({buf_id}) pybuf={pybuf}")
        self.buffers[buf_id] = pybuf

    def request_update(self, buf_id) -> int:
        cdef bint update = evdi_request_update(self.handle, buf_id)
        log(f"evdi_request_update(%#x, {buf_id})={update}", <uintptr_t> self.handle)
        return update

    def unregister_buffers(self) -> None:
        for buf_id in self.buffers.keys():
            log(f"unregister_buffer {buf_id}")
            evdi_unregister_buffer(self.handle, buf_id)
        self.buffers = {}
        self.export_buffer = 0

    def cleanup(self) -> None:
        self.unregister_buffers()
        self.disconnect()
        self.close()

    def __del__(self):
        self.cleanup()


def test_device(int device) -> bool:
    log(f"opening card {device}")
    cdef EvdiDevice d = EvdiDevice(device)
    d.open()
    d.connect(DEFAULT_EDID)
    d.enable_cursor_events()
    d.event_loop()
    d.cleanup()
    return True


def find_evdi_devices() -> List[str]:
    import os
    devices = []
    for f in sorted(os.listdir("/dev/dri")):
        if not f.startswith("card"):
            continue
        try:
            device = int(f[len("card"):])
            r = evdi_check_device(device)
            log(f"find_evdi_devices() evdi_check_device({device})={r}")
            if r==AVAILABLE:
                devices.append(device)
        except ValueError:
            pass
    log(f"find_evdi_devices()={devices}")
    return devices


def add_evdi_device() -> None:
    r = evdi_add_device()
    log("evdi_add_device()=%i", r)


def selftest(full=False) -> None:
    from xpra.log import LOG_FORMAT, enable_color
    format_string = LOG_FORMAT
    enable_color(format_string=format_string)
    log("evdi version " + ".".join(str(x) for x in get_version()))
    if full:
        #capture_logging()
        devices = find_evdi_devices()
        if devices:
            for device in devices:
                test_device(device)
        #reset_logging()
